'''
Copyright (c) 2011-2014, Agora Games, LLC All rights reserved.

https://github.com/agoragames/haigha/blob/master/LICENSE.txt
'''

from collections import deque

from haigha.writer import Writer
from haigha.classes import ProtocolClass
from haigha.frames import MethodFrame

class ExchangeClass(ProtocolClass):
  '''
  Implements the AMQP Exchange class
  '''

  def __init__(self, *args, **kwargs):
    super(ExchangeClass, self).__init__(*args, **kwargs)
    self.dispatch_map = {
      11 : self._recv_declare_ok,
      21 : self._recv_delete_ok,
    }

    self._declare_cb = deque()
    self._delete_cb = deque()

  @property
  def name(self):
    return 'exchange'

  def _cleanup(self):
    '''
    Cleanup local data.
    '''
    self._declare_cb = None
    self._delete_cb = None
    super(ExchangeClass,self)._cleanup()

  def declare(self, exchange, type, passive=False, durable=False,
      nowait=True, arguments=None, ticket=None, cb=None):
    """
    Declare the exchange.

    exchange - The name of the exchange to declare
    type - One of
    """
    nowait = nowait and self.allow_nowait() and not cb

    args = Writer()
    args.write_short(ticket or self.default_ticket).\
      write_shortstr(exchange).\
      write_shortstr(type).\
      write_bits(passive, durable, False, False, nowait).\
      write_table(arguments or {})
    self.send_frame( MethodFrame(self.channel_id, 40, 10, args) )

    if not nowait:
      self._declare_cb.append( cb )
      self.channel.add_synchronous_cb( self._recv_declare_ok )

  def _recv_declare_ok(self, _method_frame):
    '''
    Confirmation that exchange was declared.
    '''
    cb = self._declare_cb.popleft()
    if cb: cb()

  def delete(self, exchange, if_unused=False, nowait=True, ticket=None, cb=None):
    '''
    Delete an exchange.
    '''
    nowait = nowait and self.allow_nowait() and not cb

    args = Writer()
    args.write_short(ticket or self.default_ticket).\
      write_shortstr(exchange).\
      write_bits(if_unused, nowait)
    self.send_frame( MethodFrame(self.channel_id, 40, 20, args) )

    if not nowait:
      self._delete_cb.append( cb )
      self.channel.add_synchronous_cb( self._recv_delete_ok )

  def _recv_delete_ok(self, _method_frame):
    '''
    Confirmation that exchange was deleted.
    '''
    cb = self._delete_cb.popleft()
    if cb: cb()



'''
Copyright (c) 2011-2014, Agora Games, LLC All rights reserved.

https://github.com/agoragames/haigha/blob/master/LICENSE.txt
'''

class ProtocolClass(object):
  '''
  The base class of all protocol classes.
  '''

  class ProtocolError(Exception): pass
  class InvalidMethod(ProtocolError): pass
  class FrameUnderflow(ProtocolError): pass

  dispatch_map = {}

  def __init__(self, channel):
    '''
    Construct this protocol class on a channel.
    '''
    # Cache the channel id so that cleanup can remove the circular channel
    # reference but id is still accessible (it's useful!)
    self._channel = channel
    self._channel_id = channel.channel_id

  @property
  def channel(self):
    return self._channel

  @property
  def channel_id(self):
    return self._channel_id

  @property
  def logger(self):
    return self._channel.logger

  @property
  def default_ticket(self):
    return 0

  @property
  def name(self):
    '''The name given this in the protocol, i.e. 'basic', 'tx', etc'''
    raise NotImplementedError('must provide a name for %s'%(self))

  def allow_nowait(self):
    '''Return True if the transport allows nowait, False otherwise.'''
    # hack: this is a function to make testing easier
    return not self._channel.connection.synchronous

  def _cleanup(self):
    '''
    "Private" call from Channel when it's shutting down so that local
    data can be cleaned up and references closed out. It's strongly
    recommended that subclasses call this /after/ doing their own cleanup .
    Note that this removes reference to both the channel and the dispatch
    map.
    '''
    self._channel = None
    self.dispatch_map = None

  def dispatch(self, method_frame):
    '''
    Dispatch a method for this protocol.
    '''
    method = self.dispatch_map.get( method_frame.method_id )
    if method:
      callback = self.channel.clear_synchronous_cb( method )
      callback(method_frame)
    else:
      raise self.InvalidMethod("no method is registered with id: %d" % method_frame.method_id)

  def send_frame(self, frame):
    '''
    Send a frame
    '''
    self.channel.send_frame( frame )



'''
Copyright (c) 2011-2014, Agora Games, LLC All rights reserved.

https://github.com/agoragames/haigha/blob/master/LICENSE.txt
'''

from chai import Chai
import struct
import time
from datetime import datetime

from haigha.frames import method_frame
from haigha.frames.method_frame import MethodFrame
from haigha.reader import Reader
from haigha.writer import Writer

class MethodFrameTest(Chai):

  def test_type(self):
    assert_equals( 1, MethodFrame.type() )

  def test_properties(self):
    frame = MethodFrame('channel_id', 'class_id', 'method_id', 'args')
    assert_equals( 'channel_id', frame.channel_id )
    assert_equals( 'class_id', frame.class_id )
    assert_equals( 'method_id', frame.method_id )
    assert_equals( 'args', frame.args )

  def test_parse(self):
    reader = mock()
    expect( reader.read_short ).returns('class_id')
    expect( reader.read_short ).returns('method_id')
    frame = MethodFrame.parse( 42, reader )

    assert_equals( 42, frame.channel_id )
    assert_equals( 'class_id', frame.class_id )
    assert_equals( 'method_id', frame.method_id )
    assert_equals( reader, frame.args )

  def test_str(self):
    frame = MethodFrame( 42, 5, 6, Reader(bytearray('hello')) )
    assert_equals( 'MethodFrame[channel: 42, class_id: 5, method_id: 6, args: \\x68\\x65\\x6c\\x6c\\x6f]', str(frame) )

    frame = MethodFrame( 42, 5, 6 )
    assert_equals( 'MethodFrame[channel: 42, class_id: 5, method_id: 6, args: None]', str(frame) )

  def test_write_frame(self):
    args = mock()
    expect( args.buffer ).returns( 'hello' )

    frame = MethodFrame(42, 5, 6, args)
    buf = bytearray()
    frame.write_frame( buf )

    reader = Reader(buf)
    assert_equals( 1, reader.read_octet() )
    assert_equals( 42, reader.read_short() )
    size = reader.read_long()
    start_pos = reader.tell()
    assert_equals( 5, reader.read_short() )
    assert_equals( 6, reader.read_short() )
    args_pos = reader.tell()
    assert_equals( 'hello', reader.read( size-(args_pos-start_pos) ) )
    assert_equals( 0xce, reader.read_octet() )



'''
Copyright (c) 2011-2014, Agora Games, LLC All rights reserved.

https://github.com/agoragames/haigha/blob/master/LICENSE.txt
'''

from struct import Struct, pack
from calendar import timegm
from datetime import datetime
from decimal import Decimal
from operator import xor

class Writer(object):
  """
  Implements writing of structured AMQP data. Buffers data directly to a
  bytearray or a buffer supplied in the constructor. The buffer must
  supply append, extend and struct.pack_into semantics.
  """

  def __init__(self, buf=None):
    if buf is not None:
      self._output_buffer = buf
    else:
      self._output_buffer = bytearray()

  def __str__(self):
    return ''.join( ['\\x%s'%(chr(c).encode('hex')) for c in self._output_buffer] )

  __repr__ = __str__

  def __eq__(self, other):
    if isinstance(other, Writer):
      return self._output_buffer == other._output_buffer
    return False

  def buffer(self):
    '''
    Get the buffer that this has written to. Returns bytearray.
    '''
    return self._output_buffer

  def write(self, s):
    """
    Write a plain Python string, with no special encoding.
    """
    self._output_buffer.extend( s )
    return self

  def write_bits(self, *args):
    '''
    Write multiple bits in a single byte field. The bits will be written in
    little-endian order, but should be supplied in big endian order. Will raise
    ValueError when more than 8 arguments are supplied.

    write_bits(True, False) => 0x02
    '''
    # Would be nice to make this a bit smarter
    if len(args) > 8:
      raise ValueError("Can only write 8 bits at a time")

    self._output_buffer.append( chr(
      reduce(lambda x,y: xor(x,args[y]<<y), xrange(len(args)), 0)) )

    return self

  def write_bit(self, b, pack=Struct('B').pack):
    '''
    Write a single bit. Convenience method for single bit args.
    '''
    self._output_buffer.append( pack(True if b else False) )
    return self

  def write_octet(self, n, pack=Struct('B').pack):
    """
    Write an integer as an unsigned 8-bit value.
    """
    if 0 <= n <= 255:
      self._output_buffer.append( pack(n) )
    else:
      raise ValueError('Octet %d out of range 0..255', n)
    return self

  def write_short(self, n, pack=Struct('>H').pack):
    """
    Write an integer as an unsigned 16-bit value.
    """
    if 0 <= n <= 0xFFFF:
      self._output_buffer.extend( pack(n) )
    else:
      raise ValueError('Short %d out of range 0..0xFFFF', n)
    return self

  def write_short_at(self, n, pos, pack_into=Struct('>H').pack_into):
    '''
    Write an unsigned 16bit value at a specific position in the buffer.
    Used for writing tables and frames.
    '''
    if 0 <= n <= 0xFFFF:
      pack_into(self._output_buffer, pos, n)
    else:
      raise ValueError('Short %d out of range 0..0xFFFF', n)
    return self

  def write_long(self, n, pack=Struct('>I').pack):
    """
    Write an integer as an unsigned 32-bit value.
    """
    if 0 <= n <= 0xFFFFFFFF:
      self._output_buffer.extend( pack(n) )
    else:
      raise ValueError('Long %d out of range 0..0xFFFFFFFF', n)
    return self

  def write_long_at(self, n, pos, pack_into=Struct('>I').pack_into):
    '''
    Write an unsigned 32bit value at a specific position in the buffer.
    Used for writing tables and frames.
    '''
    if 0 <= n <= 0xFFFFFFFF:
      pack_into(self._output_buffer, pos, n)
    else:
      raise ValueError('Long %d out of range 0..0xFFFFFFFF', n)
    return self

  def write_longlong(self, n, pack=Struct('>Q').pack):
    """
    Write an integer as an unsigned 64-bit value.
    """
    if 0 <= n <= 0xFFFFFFFFFFFFFFFF:
      self._output_buffer.extend( pack(n) )
    else:
      raise ValueError('Longlong %d out of range 0..0xFFFFFFFFFFFFFFFF', n)
    return self

  def write_shortstr(self, s):
    """
    Write a string up to 255 bytes long after encoding.  If passed
    a unicode string, encode as UTF-8.
    """
    if isinstance(s, unicode):
      s = s.encode('utf-8')
    self.write_octet( len(s) )
    self.write( s )
    return self

  def write_longstr(self, s):
    """
    Write a string up to 2**32 bytes long after encoding.  If passed
    a unicode string, encode as UTF-8.
    """
    if isinstance(s, unicode):
      s = s.encode('utf-8')
    self.write_long( len(s) )
    self.write( s )
    return self

  def write_timestamp(self, t, pack=Struct('>Q').pack):
    """
    Write out a Python datetime.datetime object as a 64-bit integer
    representing seconds since the Unix UTC epoch.
    """
    # Double check timestamp, can't imagine why it would be signed
    self._output_buffer.extend( pack(long(timegm(t.timetuple())) ))
    return self

  # NOTE: coding this to http://dev.rabbitmq.com/wiki/Amqp091Errata#section_3 and
  # NOT spec 0.9.1. It seems that Rabbit and other brokers disagree on this
  # section for now.
  def write_table(self, d):
    """
    Write out a Python dictionary made of up string keys, and values
    that are strings, signed integers, Decimal, datetime.datetime, or
    sub-dictionaries following the same constraints.
    """
    # HACK: encoding of AMQP tables is broken because it requires the length of
    # the /encoded/ data instead of the number of items.  To support streaming,
    # fiddle with cursor position, rewinding to write the real length of the
    # data.  Generally speaking, I'm not a fan of the AMQP encoding scheme, it
    # could be much faster.
    table_len_pos = len( self._output_buffer )
    self.write_long( 0 )
    table_data_pos = len(self._output_buffer )

    for key,value in d.iteritems():
      self._write_item( key, value )

    table_end_pos = len(self._output_buffer)
    table_len = table_end_pos - table_data_pos

    self.write_long_at( table_len, table_len_pos )
    return self

  def _write_item(self, key, value ):
    self.write_shortstr(key)
    self._write_field( value )

  def _write_field(self, value ):
    writer = self.field_type_map.get( type(value) )
    if writer:
      writer(self, value)
    else:
      for kls, writer in self.field_type_map.items():
        if isinstance(value, kls):
          writer(self, value)
          break
      else:
        # Write a None because we've already written a key
        self._field_none( value )

  def _field_bool(self, val, pack=Struct('B').pack):
    self._output_buffer.append( 't' )
    self._output_buffer.append( pack( True if val else False ) )

  def _field_int(self, val, short_pack=Struct('>h').pack, \
  int_pack=Struct('>i').pack, long_pack=Struct('>q').pack):
    if -2**15 <= val < 2**15:
      self._output_buffer.append( 's' )
      self._output_buffer.extend( short_pack(val) )
    elif -2**31 <= val < 2**31:
      self._output_buffer.append( 'I' )
      self._output_buffer.extend( int_pack(val) )
    else:
      self._output_buffer.append( 'l' )
      self._output_buffer.extend( long_pack(val) )

  def _field_double(self, val, pack=Struct('>d').pack):
    self._output_buffer.append( 'd' )
    self._output_buffer.extend( pack(val) )

  # Coding to http://dev.rabbitmq.com/wiki/Amqp091Errata#section_3 which
  # differs from spec in that the value is signed.
  def _field_decimal(self, val, exp_pack=Struct('B').pack, dig_pack=Struct('>i').pack):
    self._output_buffer.append('D')
    sign, digits, exponent = val.as_tuple()
    v = 0
    for d in digits:
      v = (v * 10) + d
    if sign:
      v = -v
    self._output_buffer.append( exp_pack(-exponent) )
    self._output_buffer.extend( dig_pack(v) )

  def _field_str(self, val):
    self._output_buffer.append('S')
    self.write_longstr(val)

  def _field_unicode(self, val):
    val = val.encode('utf-8')
    self._output_buffer.append('S')
    self.write_longstr( val )

  def _field_timestamp(self, val):
    self._output_buffer.append( 'T' )
    self.write_timestamp( val )

  def _field_table(self, val):
    self._output_buffer.append('F')
    self.write_table( val )

  def _field_none(self, val):
    self._output_buffer.append('V')

  def _field_bytearray(self, val):
    self._output_buffer.append('x')
    self.write_longstr( val )

  def _field_iterable(self, val):
    self._output_buffer.append( 'A' )
    for x in val:
      self._write_field( x )

  field_type_map = {
    bool      : _field_bool,
    int       : _field_int,
    long      : _field_int,
    float     : _field_double,
    Decimal   : _field_decimal,
    str       : _field_str,
    unicode   : _field_unicode,
    datetime  : _field_timestamp,
    dict      : _field_table,
    type(None): _field_none,
    bytearray : _field_bytearray,
  }

  # 0.9.1 spec mapping
  # field_type_map = {
  #   bool      : _field_bool,
  #   int       : _field_int,
  #   long      : _field_int,
  #   float     : _field_double,
  #   Decimal   : _field_decimal,
  #   str       : _field_str,
  #   unicode   : _field_unicode,
  #   datetime  : _field_timestamp,
  #   dict      : _field_table,
  #   None      : _field_none,
  #   bytearray : _field_bytearray,
  #   list      : _field_iterable,
  #   tuple     : _field_iterable,
  #   set       : _field_iterable,
  # }



'''
Copyright (c) 2011-2014, Agora Games, LLC All rights reserved.

https://github.com/agoragames/haigha/blob/master/LICENSE.txt
'''
from collections import deque

from chai import Chai

from haigha.classes import exchange_class, ProtocolClass, ExchangeClass
from haigha.frames import MethodFrame
from haigha.writer import Writer

class ExchangeClassTest(Chai):

  def setUp(self):
    super(ExchangeClassTest,self).setUp()
    ch = mock()
    ch.channel_id = 42
    ch.logger = mock()
    self.klass = ExchangeClass( ch )

  def test_init(self):
    expect(ProtocolClass.__init__).args('foo', a='b' )

    klass = ExchangeClass.__new__(ExchangeClass)
    klass.__init__('foo', a='b')

    assert_equals(
      {
        11 : klass._recv_declare_ok,
        21 : klass._recv_delete_ok,
      }, klass.dispatch_map )
    assert_equals( deque(), klass._declare_cb )
    assert_equals( deque(), klass._delete_cb )

  def test_cleanup(self):
    self.klass._cleanup()
    assert_equals( None, self.klass._declare_cb )
    assert_equals( None, self.klass._delete_cb )
    assert_equals( None, self.klass._channel )
    assert_equals( None, self.klass.dispatch_map )

  def test_declare_default_args(self):
    w = mock()
    expect( self.klass.allow_nowait ).returns( True )
    expect( mock(exchange_class, 'Writer') ).returns( w )
    expect( w.write_short ).args( self.klass.default_ticket ).returns(w)
    expect( w.write_shortstr ).args( 'exchange' ).returns( w )
    expect( w.write_shortstr ).args( 'topic' ).returns( w )
    expect( w.write_bits ).args( False, False, False, False, True ).returns( w )
    expect( w.write_table ).args( {} )
    expect( mock(exchange_class, 'MethodFrame') ).args(42, 40, 10, w).returns( 'frame' )
    expect( self.klass.send_frame ).args( 'frame' )
    stub( self.klass.channel.add_synchronous_cb )

    self.klass.declare('exchange', 'topic')
    assert_equals( deque(), self.klass._declare_cb )

  def test_declare_with_args(self):
    w = mock()
    stub( self.klass.allow_nowait )
    expect( mock(exchange_class, 'Writer') ).returns( w )
    expect( w.write_short ).args( 't' ).returns(w)
    expect( w.write_shortstr ).args( 'exchange' ).returns( w )
    expect( w.write_shortstr ).args( 'topic' ).returns( w )
    expect( w.write_bits ).args( 'p', 'd', False, False, False ).returns( w )
    expect( w.write_table ).args( 'table' )
    expect( mock(exchange_class, 'MethodFrame') ).args(42, 40, 10, w).returns( 'frame' )
    expect( self.klass.send_frame ).args( 'frame' )
    expect( self.klass.channel.add_synchronous_cb ).args( self.klass._recv_declare_ok )

    self.klass.declare('exchange', 'topic', passive='p', durable='d',
      nowait=False, arguments='table', ticket='t')
    assert_equals( deque([None]), self.klass._declare_cb )

  def test_declare_with_cb(self):
    w = mock()
    expect( self.klass.allow_nowait ).returns( True )
    expect( mock(exchange_class, 'Writer') ).returns( w )
    expect( w.write_short ).args( 't' ).returns(w)
    expect( w.write_shortstr ).args( 'exchange' ).returns( w )
    expect( w.write_shortstr ).args( 'topic' ).returns( w )
    expect( w.write_bits ).args( 'p', 'd', False, False, False ).returns( w )
    expect( w.write_table ).args( 'table' )
    expect( mock(exchange_class, 'MethodFrame') ).args(42, 40, 10, w).returns( 'frame' )
    expect( self.klass.send_frame ).args( 'frame' )
    expect( self.klass.channel.add_synchronous_cb ).args( self.klass._recv_declare_ok )

    self.klass.declare('exchange', 'topic', passive='p', durable='d',
      nowait=True, arguments='table', ticket='t', cb='foo')
    assert_equals( deque(['foo']), self.klass._declare_cb )

  def test_recv_declare_ok_no_cb(self):
    self.klass._declare_cb = deque([None])
    self.klass._recv_declare_ok('frame')
    assert_equals( deque(), self.klass._declare_cb )

  def test_recv_declare_ok_with_cb(self):
    cb = mock()
    self.klass._declare_cb = deque([cb])
    expect( cb )
    self.klass._recv_declare_ok('frame')
    assert_equals( deque(), self.klass._declare_cb )

  def test_delete_default_args(self):
    w = mock()
    expect( self.klass.allow_nowait ).returns( True )
    expect( mock(exchange_class, 'Writer') ).returns( w )
    expect( w.write_short ).args( self.klass.default_ticket ).returns(w)
    expect( w.write_shortstr ).args( 'exchange' ).returns( w )
    expect( w.write_bits ).args( False, True )
    expect( mock(exchange_class, 'MethodFrame') ).args(42, 40, 20, w).returns( 'frame' )
    expect( self.klass.send_frame ).args( 'frame' )
    stub( self.klass.channel.add_synchronous_cb )

    self.klass.delete('exchange')
    assert_equals( deque(), self.klass._delete_cb )

  def test_delete_with_args(self):
    w = mock()
    stub( self.klass.allow_nowait )
    expect( mock(exchange_class, 'Writer') ).returns( w )
    expect( w.write_short ).args( 't' ).returns(w)
    expect( w.write_shortstr ).args( 'exchange' ).returns( w )
    expect( w.write_bits ).args( 'maybe', False )
    expect( mock(exchange_class, 'MethodFrame') ).args(42, 40, 20, w).returns( 'frame' )
    expect( self.klass.send_frame ).args( 'frame' )
    expect( self.klass.channel.add_synchronous_cb ).args( self.klass._recv_delete_ok )

    self.klass.delete('exchange', if_unused='maybe', nowait=False, ticket='t')
    assert_equals( deque([None]), self.klass._delete_cb )

  def test_delete_with_cb(self):
    w = mock()
    expect( self.klass.allow_nowait ).returns( True )
    expect( mock(exchange_class, 'Writer') ).returns( w )
    expect( w.write_short ).args( 't' ).returns(w)
    expect( w.write_shortstr ).args( 'exchange' ).returns( w )
    expect( w.write_bits ).args( 'maybe', False )
    expect( mock(exchange_class, 'MethodFrame') ).args(42, 40, 20, w).returns( 'frame' )
    expect( self.klass.send_frame ).args( 'frame' )
    expect( self.klass.channel.add_synchronous_cb ).args( self.klass._recv_delete_ok )

    self.klass.delete('exchange', if_unused='maybe', nowait=True, ticket='t', cb='foo')
    assert_equals( deque(['foo']), self.klass._delete_cb )

  def test_recv_delete_ok_no_cb(self):
    self.klass._delete_cb = deque([None])
    self.klass._recv_delete_ok('frame')
    assert_equals( deque(), self.klass._delete_cb )

  def test_recv_delete_ok_with_cb(self):
    cb = mock()
    self.klass._delete_cb = deque([cb])
    expect( cb )
    self.klass._recv_delete_ok('frame')
    assert_equals( deque(), self.klass._delete_cb )



'''
Copyright (c) 2011-2014, Agora Games, LLC All rights reserved.

https://github.com/agoragames/haigha/blob/master/LICENSE.txt
'''

from haigha.transports import Transport

import errno
import socket

class SocketTransport(Transport):
  '''
  A simple blocking socket transport.
  '''

  def __init__(self, *args):
    super(SocketTransport,self).__init__(*args)
    self._synchronous = True
    self._buffer = bytearray()

  ###
  ### Transport API
  ###
  def connect(self, (host,port), klass=socket.socket):
    '''
    Connect assuming a host and port tuple.
    '''
    self._host = "%s:%s"%(host,port)
    self._sock = klass()
    self._sock.setblocking( True )
    self._sock.settimeout( self.connection._connect_timeout )
    if self.connection._sock_opts:
      for k,v in self.connection._sock_opts.iteritems():
        family,type = k
        self._sock.setsockopt(family, type, v)
    self._sock.connect( (host,port) )

    # After connecting, switch to full-blocking mode.
    self._sock.settimeout( None )

  def read(self, timeout=None):
    '''
    Read from the transport. If timeout>0, will only block for `timeout`
    seconds.
    '''
    if not hasattr(self,'_sock'):
      return None

    try:
      # Note that we ignore both None and 0, i.e. we either block with a
      # timeout or block completely and let gevent sort it out.
      if timeout:
        self._sock.settimeout( timeout )
      else:
        self._sock.settimeout( None )
      data = self._sock.recv( self._sock.getsockopt(socket.SOL_SOCKET,socket.SO_RCVBUF) )

      if len(data):
        if self.connection.debug > 1:
          self.connection.logger.debug( 'read %d bytes from %s'%(len(data), self._host) )
        if len(self._buffer):
          self._buffer.extend( data )
          data = self._buffer
          self._buffer = bytearray()
        return data

      # Note that no data means the socket is closed and we'll mark that
      # below

    except socket.timeout as e:
      # Note that this is implemented differently and though it would be
      # caught as an EnvironmentError, it has no errno. Not sure whose
      # fault that is.
      return None

    except EnvironmentError as e:
      # thrown if we have a timeout and no data
      if e.errno in (errno.EAGAIN,errno.EWOULDBLOCK):
        return None

      self.connection.logger.exception( 'error reading from %s'%(self._host) )

    self.connection.transport_closed( msg='error reading from %s'%(self._host) )

  def buffer(self, data):
    '''
    Buffer unused bytes from the input stream.
    '''
    if not hasattr(self,'_sock'):
      return None

    # data will always be a byte array
    if len(self._buffer):
      self._buffer.extend( data )
    else:
      self._buffer = bytearray(data)

  def write(self, data):
    '''
    Write some bytes to the transport.
    '''
    if not hasattr(self,'_sock'):
      return None

    try:
      self._sock.sendall( data )

      if self.connection.debug > 1:
        self.connection.logger.debug( 'sent %d bytes to %s'%(len(data), self._host) )

      return
    except EnvironmentError as e:
      # sockets raise this type of error, and since if sendall() fails we're
      # left in an indeterminate state, assume that any error we catch means
      # that the connection is dead. Note that this assumption requires this
      # to be a blocking socket; if we ever support non-blocking in this class
      # then this whole method has to change a lot.
      self.connection.logger.exception( 'error writing to %s'%(self._host) )

    self.connection.transport_closed( msg='error writing to %s'%(self._host) )

  def disconnect(self):
    '''
    Disconnect from the transport. Typically socket.close(). This call is
    welcome to raise exceptions, which the Connection will catch.
    '''
    if not hasattr(self,'_sock'):
      return None

    try:
      self._sock.close()
    finally:
      self._sock = None



'''
Copyright (c) 2011-2014, Agora Games, LLC All rights reserved.

https://github.com/agoragames/haigha/blob/master/LICENSE.txt
'''

class ConnectionError(Exception): '''Base class for all connection errors.'''
class ConnectionClosed(ConnectionError): '''The connection is closed.  Fatal.'''

class ChannelError(Exception): '''Base class for all channel errors.'''
class ChannelClosed(ChannelError): '''The channel is closed. Fatal.'''



'''
Copyright (c) 2011-2014, Agora Games, LLC All rights reserved.

https://github.com/agoragames/haigha/blob/master/LICENSE.txt
'''

from collections import deque

from haigha.connection import Connection
from haigha.classes import *
from haigha.writer import Writer
from haigha.frames import MethodFrame

class RabbitConnection(Connection):
  '''
  A connection specific to RabbitMQ that supports its extensions.
  '''

  def __init__(self, **kwargs):
    '''
    Initialize the connection
    '''
    class_map = kwargs.get('class_map',{}).copy()
    class_map.setdefault(40, RabbitExchangeClass)
    class_map.setdefault(60, RabbitBasicClass)
    class_map.setdefault(85, RabbitConfirmClass)
    kwargs['class_map'] = class_map

    super(RabbitConnection,self).__init__(**kwargs)

class RabbitExchangeClass(ExchangeClass):
  '''
  Exchange class Rabbit extensions
  '''

  def __init__(self, *args, **kwargs):
    super(RabbitExchangeClass,self).__init__(*args, **kwargs)
    self.dispatch_map[31] = self._recv_bind_ok
    self.dispatch_map[51] = self._recv_unbind_ok

    self._bind_cb = deque()
    self._unbind_cb = deque()

  # I hate the code copying here. Probably a better solution, like
  # functools.
  def declare(self, exchange, type, passive=False, durable=False,
      auto_delete=True, internal=False, nowait=True,
      arguments=None, ticket=None, cb=None):
    """
    Declare the exchange.

    exchange - The name of the exchange to declare
    type - One of
    """
    nowait = nowait and self.allow_nowait() and not cb

    args = Writer()
    args.write_short(ticket or self.default_ticket).\
      write_shortstr(exchange).\
      write_shortstr(type).\
      write_bits(passive, durable, auto_delete, internal, nowait).\
      write_table(arguments or {})
    self.send_frame( MethodFrame(self.channel_id, 40, 10, args) )

    if not nowait:
      self._declare_cb.append( cb )
      self.channel.add_synchronous_cb( self._recv_declare_ok )

  def bind(self, exchange, source, routing_key='', nowait=True,
      arguments={}, ticket=None, cb=None):
    '''
    Bind an exchange to another.
    '''
    nowait = nowait and self.allow_nowait() and not cb

    args = Writer()
    args.write_short(ticket or self.default_ticket).\
      write_shortstr(exchange).\
      write_shortstr(source).\
      write_shortstr(routing_key).\
      write_bit(nowait).\
      write_table(arguments or {})
    self.send_frame( MethodFrame(self.channel_id, 40, 30, args) )

    if not nowait:
      self._bind_cb.append( cb )
      self.channel.add_synchronous_cb( self._recv_bind_ok )

  def _recv_bind_ok(self, _method_frame):
    '''Confirm exchange bind.'''
    cb = self._bind_cb.popleft()
    if cb: cb()

  def unbind(self, exchange, source, routing_key='', nowait=True,
      arguments={}, ticket=None, cb=None):
    '''
    Unbind an exchange from another.
    '''
    nowait = nowait and self.allow_nowait() and not cb

    args = Writer()
    args.write_short(ticket or self.default_ticket).\
      write_shortstr(exchange).\
      write_shortstr(source).\
      write_shortstr(routing_key).\
      write_bit(nowait).\
      write_table(arguments or {})
    self.send_frame( MethodFrame(self.channel_id, 40, 40, args) )

    if not nowait:
      self._unbind_cb.append( cb )
      self.channel.add_synchronous_cb( self._recv_unbind_ok )

  def _recv_unbind_ok(self, _method_frame):
    '''Confirm exchange unbind.'''
    cb = self._unbind_cb.popleft()
    if cb: cb()

class RabbitBasicClass(BasicClass):
  '''
  Support Rabbit extensions to Basic class.
  '''

  def __init__(self, *args, **kwargs):
    super(RabbitBasicClass,self).__init__(*args, **kwargs)
    self.dispatch_map[80] = self._recv_ack
    self.dispatch_map[120] = self._recv_nack

    self._ack_listener = None
    self._nack_listener = None

    self._msg_id = 0
    self._last_ack_id = 0

  def set_ack_listener(self, cb):
    '''
    Set a callback for ack listening, to be used when the channel is
    in publisher confirm mode. Will be called with a single integer
    argument which is the id of the message as returned from publish().

    cb(message_id)
    '''
    self._ack_listener = cb

  def set_nack_listener(self, cb):
    '''
    Set a callbnack for nack listening, to be used when the channel is
    in publisher confirm mode. Will be called with a single integer
    argument which is the id of the message as returned from publish()
    and a boolean flag indicating if it can be requeued.

    cb(message_id, reque)
    '''
    self._nack_listener = cb

  # Probably a better solution here, like functools
  def publish(self, *args, **kwargs):
    '''
    Publish a message. Will return the id of the message if publisher
    confirmations are enabled, else will return 0.
    '''
    if self.channel.confirm._enabled:
      self._msg_id += 1
    super(RabbitBasicClass,self).publish(*args, **kwargs)
    return self._msg_id

  def _recv_ack(self, method_frame):
    '''Receive an ack from the broker.'''
    if self._ack_listener:
      delivery_tag = method_frame.args.read_longlong()
      multiple = method_frame.args.read_bit()
      if multiple:
        while self._last_ack_id < delivery_tag:
          self._last_ack_id += 1
          self._ack_listener(self._last_ack_id)
      else:
        self._last_ack_id = delivery_tag
        self._ack_listener(self._last_ack_id)

  def nack(self, delivery_tag, multiple=False, requeue=False):
    '''Send a nack to the broker.'''
    args = Writer()
    args.write_longlong(delivery_tag).\
      write_bits(multiple, requeue)

    self.send_frame( MethodFrame(self.channel_id, 60, 120, args) )

  def _recv_nack(self, method_frame):
    '''Receive a nack from the broker.'''
    if self._nack_listener:
      delivery_tag = method_frame.args.read_longlong()
      multiple, requeue = method_frame.args.read_bits(2)
      if multiple:
        while self._last_ack_id < delivery_tag:
          self._last_ack_id += 1
          self._nack_listener(self._last_ack_id, requeue)
      else:
        self._last_ack_id = delivery_tag
        self._nack_listener(self._last_ack_id, requeue)

class RabbitConfirmClass(ProtocolClass):
  '''
  Implementation of Rabbit's confirm class.
  '''

  def __init__(self, *args, **kwargs):
    super(RabbitConfirmClass, self).__init__(*args, **kwargs)
    self.dispatch_map = {
      11 : self._recv_select_ok,
    }

    self._enabled = False
    self._select_cb = deque()

  @property
  def name(self):
    return 'confirm'

  def select(self, nowait=True, cb=None):
    '''
    Set this channel to use publisher confirmations.
    '''
    nowait = nowait and self.allow_nowait() and not cb

    if not self._enabled:
      self._enabled = True
      self.channel.basic._msg_id = 0
      self.channel.basic._last_ack_id = 0
      args = Writer()
      args.write_bit(nowait)

      self.send_frame( MethodFrame(self.channel_id, 85, 10, args) )

      if not nowait:
        self._select_cb.append(cb)
        self.channel.add_synchronous_cb( self._recv_select_ok )

  def _recv_select_ok(self, _method_frame):
    cb = self._select_cb.popleft()
    if cb: cb()



'''
Copyright (c) 2011-2014, Agora Games, LLC All rights reserved.

https://github.com/agoragames/haigha/blob/master/LICENSE.txt
'''

from haigha.transports.socket_transport import SocketTransport

import errno
try:
  import gevent
  try:
    # Semaphore moved here since gevent-1.0b2
    from gevent.lock import Semaphore
  except ImportError:
    from gevent.coros import Semaphore
  from gevent import socket
  from gevent import pool
except ImportError:
  print 'Failed to load gevent modules'
  gevent = None
  Semaphore = None
  socket = None
  pool = None


class GeventTransport(SocketTransport):
  '''
  Transport using gevent backend. It relies on gevent's implementation of
  sendall to send whole frames at a time. On the input side, it uses a gevent
  semaphore to ensure exclusive access to the socket and input buffer.

  NOTE:
  This is new to haigha and there may be integration issues with some versions
  of RabbitMQ. In particular, what may happen when a blocking call to send a
  frame allows another thread to queue another frame, but the frames can't be
  interlaced due to how the protocol is defined or implemented. A lot of the
  'synchronous' calls have very specific expectations. If this becomes a
  problem then the quickest way to solve it would be to switch to non-blocking
  so that the same thread of execution is allowed to send all of its frames in
  sequence. In that case, the gevent implementation may be pushed into the
  EventSocket as another supported concurrency lib.

  Note also that the blocking nature of the sockets means that the any threads
  running IO should actively yield with a sleep(0) to ensure other threads are
  serviced. In a saturated environment, failure to do so will lead to
  significant lags in signal handling or other IO. In practice, a typical
  client is attaching to other data stores so those could be enough to yield.
  '''

  def __init__(self, *args, **kwargs):
    super(GeventTransport,self).__init__(*args)

    self._synchronous = kwargs.get('synchronous',False)
    self._read_lock = Semaphore()
    self._write_lock = Semaphore()

  ###
  ### Transport API
  ###

  def connect(self, (host,port)):
    '''
    Connect using a host,port tuple
    '''
    super(GeventTransport,self).connect( (host,port), klass=socket.socket )

  def read(self, timeout=None):
    '''
    Read from the transport. If no data is available, should return None. If
    timeout>0, will only block for `timeout` seconds.
    '''
    self._read_lock.acquire()
    try:
      return super(GeventTransport,self).read(timeout=timeout)
    finally:
      self._read_lock.release()

  def buffer(self, data):
    '''
    Buffer unused bytes from the input stream.
    '''
    self._read_lock.acquire()
    try:
      return super(GeventTransport,self).buffer(data)
    finally:
      self._read_lock.release()

  def write(self, data):
    '''
    Write some bytes to the transport.
    '''
    # MUST use a lock here else gevent could raise an exception if 2 greenlets
    # try to write at the same time. I was hoping that sendall() would do that
    # blocking for me, but I guess not. May require an eventsocket-like buffer
    # to speed up under high load.
    self._write_lock.acquire()
    try:
      return super(GeventTransport,self).write(data)
    finally:
      self._write_lock.release()


class GeventPoolTransport(GeventTransport):

  def __init__(self, *args, **kwargs):
    super(GeventPoolTransport,self).__init__(*args)

    self._pool = kwargs.get('pool',None)
    if not self._pool:
      self._pool = gevent.pool.Pool()

  @property
  def pool(self):
    '''Get a handle to the gevent pool.'''
    return self._pool

  def process_channels(self, channels):
    '''
    Process a set of channels by calling Channel.process_frames() on each.
    Some transports may choose to do this in unique ways, such as through
    a pool of threads.

    The default implementation will simply iterate over them and call
    process_frames() on each.
    '''
    for channel in channels:
      self._pool.spawn( channel.process_frames )



'''
Copyright (c) 2011-2014, Agora Games, LLC All rights reserved.

https://github.com/agoragames/haigha/blob/master/LICENSE.txt
'''

from chai import Chai

from haigha.classes import basic_class, ProtocolClass, BasicClass
from haigha.frames import MethodFrame
from haigha.writer import Writer
from haigha.reader import Reader
from haigha.message import Message
from haigha.connection import Connection

from collections import deque

class BasicClassTest(Chai):

  def setUp(self):
    super(BasicClassTest,self).setUp()
    ch = mock()
    ch.channel_id = 42
    ch.logger = mock()
    self.klass = BasicClass( ch )

  def test_init(self):
    expect(ProtocolClass.__init__).args('foo', a='b' )

    klass = BasicClass.__new__(BasicClass)
    klass.__init__('foo', a='b')

    assert_equals(
      {
        11 : klass._recv_qos_ok,
        21 : klass._recv_consume_ok,
        31 : klass._recv_cancel_ok,
        50 : klass._recv_return,
        60 : klass._recv_deliver,
        71 : klass._recv_get_response,
        72 : klass._recv_get_response,
        111 : klass._recv_recover_ok,
      }, klass.dispatch_map )
    assert_equals( 0, klass._consumer_tag_id )
    assert_equals( deque(), klass._pending_consumers )
    assert_equals( {}, klass._consumer_cb )
    assert_equals( deque(), klass._get_cb )
    assert_equals( deque(), klass._recover_cb )
    assert_equals( deque(), klass._cancel_cb )

  def test_cleanup(self):
    self.klass._cleanup()
    assert_equals( None, self.klass._pending_consumers )
    assert_equals( None, self.klass._consumer_cb )
    assert_equals( None, self.klass._get_cb )
    assert_equals( None, self.klass._recover_cb )
    assert_equals( None, self.klass._cancel_cb )
    assert_equals( None, self.klass._channel )
    assert_equals( None, self.klass.dispatch_map )

  def test_generate_consumer_tag(self):
    assert_equals( 0, self.klass._consumer_tag_id )
    assert_equals( 'channel-42-1', self.klass._generate_consumer_tag() )
    assert_equals( 1, self.klass._consumer_tag_id )

  def test_qos_default_args(self):
    w = mock()
    expect( mock( basic_class, 'Writer' ) ).returns( w )
    expect( w.write_long ).args( 0 ).returns( w )
    expect( w.write_short ).args( 0 ).returns( w )
    expect( w.write_bit ).args( False )
    expect( mock( basic_class, 'MethodFrame' ) ).args( 42, 60, 10, w ).returns( 'frame' )
    expect( self.klass.send_frame ).args( 'frame' )
    expect( self.klass.channel.add_synchronous_cb ).args( self.klass._recv_qos_ok )

    self.klass.qos()

  def test_qos_with_args(self):
    w = mock()
    expect( mock( basic_class, 'Writer' ) ).returns( w )
    expect( w.write_long ).args( 1 ).returns( w )
    expect( w.write_short ).args( 2 ).returns( w )
    expect( w.write_bit ).args( 3 )
    expect( mock( basic_class, 'MethodFrame' ) ).args( 42, 60, 10, w ).returns( 'frame' )
    expect( self.klass.send_frame ).args( 'frame' )
    expect( self.klass.channel.add_synchronous_cb ).args( self.klass._recv_qos_ok )

    self.klass.qos(prefetch_size=1, prefetch_count=2, is_global=3)

  def test_recv_qos_ok(self):
    self.klass._recv_qos_ok( 'frame' )

  def test_consume_default_args(self):
    w = mock()
    expect( self.klass.allow_nowait ).returns( True )
    expect( self.klass._generate_consumer_tag ).returns( 'ctag' )
    expect( mock(basic_class, 'Writer') ).returns( w )
    expect( w.write_short ).args( self.klass.default_ticket ).returns( w )
    expect( w.write_shortstr ).args( 'queue' ).returns( w )
    expect( w.write_shortstr ).args( 'ctag' ).returns( w )
    expect( w.write_bits ).args( False, True, False, True ).returns( w )
    expect( w.write_table ).args( {} )
    expect( mock(basic_class,'MethodFrame') ).args(42, 60, 20, w).returns( 'frame' )
    expect( self.klass.send_frame ).args( 'frame' )

    assert_equals( deque(), self.klass._pending_consumers )
    assert_equals( {}, self.klass._consumer_cb )
    self.klass.consume( 'queue', 'consumer' )
    assert_equals( deque(), self.klass._pending_consumers )
    assert_equals( {'ctag':'consumer'}, self.klass._consumer_cb )

  def test_consume_with_args_including_nowait_and_ticket(self):
    w = mock()
    stub( self.klass.allow_nowait )
    expect( mock(basic_class, 'Writer') ).returns( w )
    expect( w.write_short ).args( 'train' ).returns( w )
    expect( w.write_shortstr ).args( 'queue' ).returns( w )
    expect( w.write_shortstr ).args( 'stag' ).returns( w )
    expect( w.write_bits ).args( 'nloc', 'nack', 'mine', False ).returns( w )
    expect( w.write_table ).args( {} )
    expect( mock(basic_class,'MethodFrame') ).args(42, 60, 20, w).returns( 'frame' )
    expect( self.klass.send_frame ).args( 'frame' )
    expect( self.klass.channel.add_synchronous_cb ).args( self.klass._recv_consume_ok )

    assert_equals( deque(), self.klass._pending_consumers )
    assert_equals( {}, self.klass._consumer_cb )
    self.klass.consume( 'queue', 'consumer', consumer_tag='stag', no_local='nloc',
      no_ack='nack', exclusive='mine', nowait=False, ticket='train' )
    assert_equals( deque([('consumer',None)]), self.klass._pending_consumers )
    assert_equals( {}, self.klass._consumer_cb )

  def test_consume_with_args_including_nowait_no_ticket_with_callback(self):
    w = mock()
    stub( self.klass._generate_consumer_tag )
    expect( mock(basic_class, 'Writer') ).returns( w )
    expect( w.write_short ).args( self.klass.default_ticket ).returns( w )
    expect( w.write_shortstr ).args( 'queue' ).returns( w )
    expect( w.write_shortstr ).args( 'stag' ).returns( w )
    expect( w.write_bits ).args( 'nloc', 'nack', 'mine', False ).returns( w )
    expect( w.write_table ).args( {} )
    expect( mock(basic_class,'MethodFrame') ).args(42, 60, 20, w).returns( 'frame' )
    expect( self.klass.send_frame ).args( 'frame' )
    expect( self.klass.channel.add_synchronous_cb ).args( self.klass._recv_consume_ok )

    self.klass._pending_consumers = deque([('blargh',None)])
    assert_equals( {}, self.klass._consumer_cb )
    self.klass.consume( 'queue', 'consumer', consumer_tag='stag', no_local='nloc',
      no_ack='nack', exclusive='mine', nowait=False, cb='callback' )
    assert_equals( deque([('blargh',None),('consumer','callback')]), self.klass._pending_consumers )
    assert_equals( {}, self.klass._consumer_cb )

  def test_recv_consume_ok(self):
    frame = mock()
    cb = mock()
    expect( frame.args.read_shortstr ).returns( 'ctag' )
    self.klass._pending_consumers = deque([('consumer',None), ('blargh',cb)])

    assert_equals( {}, self.klass._consumer_cb )
    self.klass._recv_consume_ok( frame )
    assert_equals( {'ctag':'consumer'}, self.klass._consumer_cb )
    assert_equals( deque([('blargh',cb)]), self.klass._pending_consumers )

    # call again and assert that cb is called
    frame2 = mock()
    expect( frame2.args.read_shortstr ).returns( 'ctag2' )
    expect( cb )
    self.klass._recv_consume_ok( frame2 )
    assert_equals( {'ctag':'consumer','ctag2':'blargh'}, self.klass._consumer_cb )
    assert_equals( deque(), self.klass._pending_consumers )

  def test_cancel_default_args(self):
    w = mock()
    expect( self.klass.allow_nowait ).returns( True )
    expect( mock(basic_class, 'Writer') ).returns( w )
    expect( w.write_shortstr ).args( '' ).returns( w )
    expect( w.write_bit ).args( True )

    expect( mock(basic_class,'MethodFrame') ).args(42, 60, 30, w).returns( 'frame' )
    expect( self.klass.send_frame ).args( 'frame' )

    self.klass._consumer_cb[ '' ] = 'foo'
    assert_equals( deque(), self.klass._cancel_cb )
    self.klass.cancel()
    assert_equals( deque(), self.klass._cancel_cb )
    assert_equals( {}, self.klass._consumer_cb )

  def test_cancel_nowait_and_consumer_tag_not_registered(self):
    w = mock()
    expect( self.klass.allow_nowait ).returns( True )
    expect( mock(basic_class, 'Writer') ).returns( w )
    expect( w.write_shortstr ).args( 'ctag' ).returns( w )
    expect( w.write_bit ).args( True )

    expect( mock(basic_class,'MethodFrame') ).args(42, 60, 30, w).returns( 'frame' )
    expect( self.klass.send_frame ).args( 'frame' )

    expect( self.klass.logger.warning ).args(
      'no callback registered for consumer tag " %s "', 'ctag' )

    assert_equals( deque(), self.klass._cancel_cb )
    self.klass.cancel( consumer_tag='ctag' )
    assert_equals( deque(), self.klass._cancel_cb )
    assert_equals( {}, self.klass._consumer_cb )

  def test_cancel_wait_without_cb(self):
    w = mock()
    stub( self.klass.allow_nowait )
    expect( mock(basic_class, 'Writer') ).returns( w )
    expect( w.write_shortstr ).args( '' ).returns( w )
    expect( w.write_bit ).args( False )

    expect( mock(basic_class,'MethodFrame') ).args(42, 60, 30, w).returns( 'frame' )
    expect( self.klass.send_frame ).args( 'frame' )

    expect( self.klass.channel.add_synchronous_cb ).args( self.klass._recv_cancel_ok )

    assert_equals( deque(), self.klass._cancel_cb )
    self.klass.cancel( nowait=False )
    assert_equals( deque([None]), self.klass._cancel_cb )
    assert_equals( {}, self.klass._consumer_cb )

  def test_cancel_wait_with_cb(self):
    w = mock()
    stub( self.klass.allow_nowait )
    expect( mock(basic_class, 'Writer') ).returns( w )
    expect( w.write_shortstr ).args( '' ).returns( w )
    expect( w.write_bit ).args( False )

    expect( mock(basic_class,'MethodFrame') ).args(42, 60, 30, w).returns( 'frame' )
    expect( self.klass.send_frame ).args( 'frame' )

    expect( self.klass.channel.add_synchronous_cb ).args( self.klass._recv_cancel_ok )

    self.klass._cancel_cb = deque(['blargh'])
    self.klass.cancel( nowait=False, cb='user_cb' )
    assert_equals( deque(['blargh','user_cb']), self.klass._cancel_cb )
    assert_equals( {}, self.klass._consumer_cb )

  def test_cancel_resolves_to_ctag_when_consumer_arg_supplied(self):
    w = mock()
    expect( self.klass.allow_nowait ).returns( True )
    expect( mock(basic_class, 'Writer') ).returns( w )
    expect( w.write_shortstr ).args( 'ctag' ).returns( w )
    expect( w.write_bit ).args( True )

    expect( mock(basic_class,'MethodFrame') ).args(42, 60, 30, w).returns( 'frame' )
    expect( self.klass.send_frame ).args( 'frame' )

    self.klass._consumer_cb[ 'ctag' ] = 'consumer'
    assert_equals( deque(), self.klass._cancel_cb )
    self.klass.cancel( consumer='consumer' )
    assert_equals( deque(), self.klass._cancel_cb )
    assert_equals( {}, self.klass._consumer_cb )

  def test_recv_cancel_ok_when_consumer_and_callback(self):
    frame = mock()
    cancel_cb = mock()
    expect( frame.args.read_shortstr ).returns( 'ctag' )
    self.klass._consumer_cb['ctag'] = 'foo'
    self.klass._cancel_cb = deque([ cancel_cb, mock() ])
    expect( cancel_cb )

    self.klass._recv_cancel_ok( frame )
    assert_equals( 1,len(self.klass._cancel_cb) )
    assert_false( cancel_cb in self.klass._cancel_cb )

  def test_recv_cancel_ok_when_no_consumer_or_callback(self):
    frame = mock()
    expect( frame.args.read_shortstr ).returns( 'ctag' )
    expect( self.klass.logger.warning ).args(
      'no callback registered for consumer tag " %s "', 'ctag' )
    self.klass._cancel_cb = deque( [ None, mock() ] )

    self.klass._recv_cancel_ok( frame )
    assert_equals( 1, len(self.klass._cancel_cb) )
    assert_false( None in self.klass._cancel_cb  )

  def test_publish_default_args(self):
    args = Writer()
    msg = Message('hello, world')
    args.write_short(0)\
        .write_shortstr('exchange')\
        .write_shortstr('routing_key')\
        .write_bits(False, False)
    self.klass.channel.connection.frame_max = 3

    expect( mock(basic_class, 'MethodFrame') ).args(42, 60, 40, args).returns( 'methodframe' )
    expect( mock(basic_class, 'HeaderFrame') ).args(42, 60, 0, len(msg), msg.properties).returns( 'headerframe' )
    expect(mock(basic_class, 'ContentFrame').create_frames).args(42, msg.body, 3).returns(['f0', 'f1', 'f2'])
    expect(self.klass.send_frame).args('methodframe')
    expect(self.klass.send_frame).args('headerframe')
    expect(self.klass.send_frame).args('f0')
    expect(self.klass.send_frame).args('f1')
    expect(self.klass.send_frame).args('f2')
    self.klass.publish(msg, 'exchange', 'routing_key')

  def test_publish_with_args(self):
    w = mock()
    msg = Message('hello, world')
    expect( mock(basic_class, 'Writer') ).returns( w )
    expect( w.write_short ).args( 'ticket' ).returns( w )
    expect( w.write_shortstr ).args( 'exchange' ).returns( w )
    expect( w.write_shortstr ).args( 'route' ).returns( w )
    expect( w.write_bits ).args( 'm','i' )
    self.klass.channel.connection.frame_max = 3

    expect( mock(basic_class, 'MethodFrame') ).args(42, 60, 40, w).returns( 'methodframe' )
    expect( mock(basic_class, 'HeaderFrame') ).args(42, 60, 0, len(msg), msg.properties).returns( 'headerframe' )
    expect(mock(basic_class, 'ContentFrame').create_frames).args(42, msg.body, 3).returns(['f0', 'f1', 'f2'])
    expect(self.klass.send_frame).args('methodframe')
    expect(self.klass.send_frame).args('headerframe')
    expect(self.klass.send_frame).args('f0')
    expect(self.klass.send_frame).args('f1')
    expect(self.klass.send_frame).args('f2')

    self.klass.publish(msg, 'exchange', 'route', mandatory='m', immediate='i', ticket='ticket' )

  def test_return_msg(self):
    args = Writer()
    args.write_short(3)
    args.write_shortstr('reply_text')
    args.write_shortstr('exchange')
    args.write_shortstr('routing_key')
    expect(mock(basic_class, 'MethodFrame')).args(42, 60, 50, args).returns('frame')
    expect(self.klass.send_frame).args('frame')
    self.klass.return_msg(3, 'reply_text', 'exchange', 'routing_key')

  def test_recv_return(self):
    self.klass._recv_return( 'frame' )

  def test_recv_deliver_with_cb(self):
    msg = mock()
    msg.delivery_info = {'consumer_tag':'ctag'}
    cb = mock()
    self.klass._consumer_cb['ctag'] = cb

    expect( self.klass._read_msg ).args(
      'frame', with_consumer_tag=True, with_message_count=False ).returns( msg )
    expect( cb ).args( msg )

    self.klass._recv_deliver('frame')

  def test_recv_deliver_without_cb(self):
    msg = mock()
    msg.delivery_info = {'consumer_tag':'ctag'}

    expect( self.klass._read_msg ).args(
     'frame', with_consumer_tag=True, with_message_count=False  ).returns( msg )

    self.klass._recv_deliver('frame')

  def test_get_default_args(self):
    w = mock()
    expect( mock(basic_class, 'Writer') ).returns( w )
    expect( w.write_short ).args( self.klass.default_ticket ).returns( w )
    expect( w.write_shortstr ).args( 'queue' ).returns( w )
    expect( w.write_bit ).args( True )
    expect( mock(basic_class,'MethodFrame') ).args(42, 60, 70, w).returns( 'frame' )
    expect( self.klass.send_frame ).args( 'frame' )
    expect( self.klass.channel.add_synchronous_cb ).args(
      self.klass._recv_get_response ).returns('msg')

    assert_equals( deque(), self.klass._get_cb )
    assert_equals( 'msg', self.klass.get('queue') )
    assert_equals( deque([None]), self.klass._get_cb )

  def test_get_with_args(self):
    w = mock()
    expect( mock(basic_class, 'Writer') ).returns( w )
    expect( w.write_short ).args( 'ticket' ).returns( w )
    expect( w.write_shortstr ).args( 'queue' ).returns( w )
    expect( w.write_bit ).args( 'ack' )
    expect( mock(basic_class,'MethodFrame') ).args(42, 60, 70, w).returns( 'frame' )
    expect( self.klass.send_frame ).args( 'frame' )
    expect( self.klass.channel.add_synchronous_cb ).args(
      self.klass._recv_get_response ).returns('msg')

    self.klass._get_cb = deque(['blargh'])
    assert_equals( 'msg', self.klass.get('queue', 'consumer', no_ack='ack', ticket='ticket') )
    assert_equals( deque(['blargh','consumer']), self.klass._get_cb )

  def test_recv_get_response(self):
    frame = mock()
    frame.method_id = 71
    expect( self.klass._recv_get_ok ).args( frame ).returns('msg' )
    assert_equals( 'msg', self.klass._recv_get_response(frame) )

    frame.method_id = 72
    expect( self.klass._recv_get_empty ).args( frame ).returns('nada')
    assert_equals( 'nada', self.klass._recv_get_response(frame) )

  def test_recv_get_ok_with_cb(self):
    cb = mock()
    self.klass._get_cb.append( cb )
    self.klass._get_cb.append( mock() )

    expect( self.klass._read_msg ).args(
      'frame', with_consumer_tag=False, with_message_count=True ).returns( 'msg' )
    expect( cb ).args( 'msg' )

    assert_equals( 'msg', self.klass._recv_get_ok('frame') )
    assert_equals( 1, len(self.klass._get_cb) )
    assert_false( cb in self.klass._get_cb )

  def test_recv_get_ok_without_cb(self):
    self.klass._get_cb.append( None )
    self.klass._get_cb.append( mock() )

    expect( self.klass._read_msg ).args(
      'frame', with_consumer_tag=False, with_message_count=True ).returns( 'msg' )

    self.klass._recv_get_ok( 'frame' )
    assert_equals( 1, len(self.klass._get_cb) )
    assert_false( None in self.klass._get_cb )

  def test_recv_get_empty_with_cb(self):
    cb = mock()
    self.klass._get_cb.append( cb )
    self.klass._get_cb.append( mock() )

    expect( cb ).args( None )

    self.klass._recv_get_empty( 'frame' )
    assert_equals( 1, len(self.klass._get_cb) )
    assert_false( cb in self.klass._get_cb )

  def test_recv_get_empty_without_cb(self):
    self.klass._get_cb.append( None )
    self.klass._get_cb.append( mock() )

    self.klass._recv_get_empty( 'frame' )
    assert_equals( 1, len(self.klass._get_cb) )
    assert_false( None in self.klass._get_cb )

  def test_ack_default_args(self):
    w = mock()
    expect( mock( basic_class, 'Writer' ) ).returns( w )
    expect( w.write_longlong ).args( 8675309 ).returns( w )
    expect( w.write_bit ).args( False )
    expect( mock( basic_class, 'MethodFrame' ) ).args( 42, 60, 80, w ).returns( 'frame' )
    expect( self.klass.send_frame ).args( 'frame' )

    self.klass.ack( 8675309 )

  def test_ack_with_args(self):
    w = mock()
    expect( mock( basic_class, 'Writer' ) ).returns( w )
    expect( w.write_longlong ).args( 8675309 ).returns( w )
    expect( w.write_bit ).args( 'many' )
    expect( mock( basic_class, 'MethodFrame' ) ).args( 42, 60, 80, w ).returns( 'frame' )
    expect( self.klass.send_frame ).args( 'frame' )

    self.klass.ack( 8675309, multiple='many' )

  def test_reject_default_args(self):
    w = mock()
    expect( mock( basic_class, 'Writer' ) ).returns( w )
    expect( w.write_longlong ).args( 8675309 ).returns( w )
    expect( w.write_bit ).args( False )
    expect( mock( basic_class, 'MethodFrame' ) ).args( 42, 60, 90, w ).returns( 'frame' )
    expect( self.klass.send_frame ).args( 'frame' )

    self.klass.reject( 8675309 )

  def test_reject_with_args(self):
    w = mock()
    expect( mock( basic_class, 'Writer' ) ).returns( w )
    expect( w.write_longlong ).args( 8675309 ).returns( w )
    expect( w.write_bit ).args( 'sure' )
    expect( mock( basic_class, 'MethodFrame' ) ).args( 42, 60, 90, w ).returns( 'frame' )
    expect( self.klass.send_frame ).args( 'frame' )

    self.klass.reject( 8675309, requeue='sure' )

  def test_recover_async(self):
    w = mock()
    expect( mock( basic_class, 'Writer' ) ).returns( w )
    expect( w.write_bit ).args( False )
    expect( mock( basic_class, 'MethodFrame' ) ).args( 42, 60, 100, w ).returns( 'frame' )
    expect( self.klass.send_frame ).args( 'frame' )

    self.klass.recover_async()

  def test_recover_default_args(self):
    w = mock()
    expect( mock( basic_class, 'Writer' ) ).returns( w )
    expect( w.write_bit ).args( False )
    expect( mock( basic_class, 'MethodFrame' ) ).args( 42, 60, 110, w ).returns( 'frame' )
    expect( self.klass.send_frame ).args( 'frame' )
    expect( self.klass.channel.add_synchronous_cb ).args( self.klass._recv_recover_ok )

    self.klass.recover()
    assert_equals( deque([None]), self.klass._recover_cb )

  def test_recover_with_args(self):
    w = mock()
    expect( mock( basic_class, 'Writer' ) ).returns( w )
    expect( w.write_bit ).args( 'requeue' )
    expect( mock( basic_class, 'MethodFrame' ) ).args( 42, 60, 110, w ).returns( 'frame' )
    expect( self.klass.send_frame ).args( 'frame' )
    expect( self.klass.channel.add_synchronous_cb ).args( self.klass._recv_recover_ok )

    self.klass._recover_cb = deque(['blargh'])
    self.klass.recover(requeue='requeue', cb='callback')
    assert_equals( deque(['blargh','callback']), self.klass._recover_cb )

  def test_recv_recover_ok_with_cb(self):
    cb = mock()
    self.klass._recover_cb.append( cb )
    self.klass._recover_cb.append( mock() )

    expect( cb )

    self.klass._recv_recover_ok( 'frame' )
    assert_equals( 1, len(self.klass._recover_cb) )
    assert_false( cb in self.klass._recover_cb )

  def test_recv_recover_ok_without_cb(self):
    self.klass._recover_cb.append( None )
    self.klass._recover_cb.append( mock() )

    self.klass._recv_recover_ok( 'frame' )
    assert_equals( 1, len(self.klass._recover_cb) )
    assert_false( None in self.klass._recover_cb )

  def test_read_msg_raises_frameunderflow_when_no_header_frame(self):
    expect(self.klass.channel.next_frame).returns(None)
    expect(self.klass.channel.requeue_frames).args(['method_frame'])
    assert_raises(self.klass.FrameUnderflow, self.klass._read_msg, 'method_frame')

  def test_read_msg_raises_frameunderflow_when_no_content_frames(self):
    header_frame = mock()
    header_frame.size = 1000000
    expect(self.klass.channel.next_frame).returns(header_frame)
    expect(self.klass.channel.next_frame).returns(None)
    expect(self.klass.channel.requeue_frames).args(deque([header_frame, 'method_frame']))
    assert_raises(self.klass.FrameUnderflow, self.klass._read_msg, 'method_frame')

  def test_read_msg_when_body_length_0_no_cb(self):
    method_frame = mock()
    header_frame = mock()
    header_frame.size = 0
    header_frame.properties = {'foo':'bar'}
    delivery_info = {'channel': self.klass.channel,
                     'consumer_tag': 'consumer_tag',
                     'delivery_tag': 9,
                     'redelivered': False,
                     'exchange': 'exchange',
                     'routing_key': 'routing_key'}

    expect(self.klass.channel.next_frame).returns(header_frame)
    expect(method_frame.args.read_shortstr).returns('consumer_tag')
    expect(method_frame.args.read_longlong).returns(9)
    expect(method_frame.args.read_bit).returns(False)
    expect(method_frame.args.read_shortstr).returns('exchange')
    expect(method_frame.args.read_shortstr).returns('routing_key')
    expect(Message).args(
      body=bytearray(), delivery_info=delivery_info, foo='bar').returns('message')

    assert_equals( 'message', self.klass._read_msg(method_frame, with_consumer_tag=True) )

  def test_read_msg_when_body_length_greater_than_0_with_cb(self):
    method_frame = mock()
    header_frame = mock()
    header_frame.size = 100
    header_frame.properties = {}
    cframe1 = mock()
    cframe2 = mock()
    self.klass._consumer_cb['ctag'] = mock()
    delivery_info = {
      'channel': self.klass.channel,
      'delivery_tag': 'dtag',
      'redelivered': 'no',
      'exchange': 'exchange',
      'routing_key': 'routing_key',
      'message_count': 8675309,
    }

    expect(self.klass.channel.next_frame).returns(header_frame)
    expect(self.klass.channel.next_frame).returns(cframe1)
    expect(cframe1.payload.buffer).returns('x'*50)
    expect(self.klass.channel.next_frame).returns(cframe2)
    expect(cframe2.payload.buffer).returns('x'*50)
    expect(method_frame.args.read_longlong).returns('dtag')
    expect(method_frame.args.read_bit).returns('no')
    expect(method_frame.args.read_shortstr).returns('exchange')
    expect(method_frame.args.read_shortstr).returns('routing_key')
    expect(method_frame.args.read_long).returns(8675309)
    expect(Message).args(
      body=bytearray('x'*100), delivery_info=delivery_info).returns('message')

    assert_equals( 'message', self.klass._read_msg(method_frame, with_message_count=True) )



'''
Copyright (c) 2011-2014, Agora Games, LLC All rights reserved.

https://github.com/agoragames/haigha/blob/master/LICENSE.txt
'''

from chai import Chai
import errno
import gevent
from gevent.coros import Semaphore
from gevent import socket
from gevent.pool import Pool

from haigha.transports import gevent_transport
from haigha.transports.gevent_transport import *

class GeventTransportTest(Chai):

  def setUp(self):
    super(GeventTransportTest,self).setUp()

    self.connection = mock()
    self.transport = GeventTransport(self.connection)
    self.transport._host = 'server:1234'

  def test_init(self):
    assert_equals( bytearray(), self.transport._buffer )
    assert_false( self.transport._synchronous )
    assert_true( isinstance(self.transport._read_lock,Semaphore) )
    assert_true( isinstance(self.transport._write_lock,Semaphore) )

  def test_connect(self):
    with expect( mock(gevent_transport,'super') ).args(is_arg(GeventTransport), GeventTransport).returns(mock()) as parent:
      expect( parent.connect ).args(('host','port'), klass=is_arg(socket.socket)).returns( 'somedata' )

    self.transport.connect( ('host','port') )

  def test_read(self):
    self.transport._read_lock = mock()
    expect( self.transport._read_lock.acquire )
    with expect( mock(gevent_transport,'super') ).args(is_arg(GeventTransport), GeventTransport).returns(mock()) as parent:
      expect( parent.read ).args(timeout=None).returns( 'somedata' )
    expect( self.transport._read_lock.release )

    assert_equals( 'somedata', self.transport.read() )

  def test_read_when_raises_exception(self):
    self.transport._read_lock = mock()
    expect( self.transport._read_lock.acquire )
    with expect( mock(gevent_transport,'super') ).args(is_arg(GeventTransport), GeventTransport).returns(mock()) as parent:
      expect( parent.read ).args(timeout='5').raises(Exception('fail'))
    expect( self.transport._read_lock.release )

    assert_raises(Exception, self.transport.read, timeout='5')

  def test_buffer(self):
    self.transport._read_lock = mock()
    expect( self.transport._read_lock.acquire )
    with expect( mock(gevent_transport,'super') ).args(is_arg(GeventTransport), GeventTransport).returns(mock()) as parent:
      expect( parent.buffer ).args('datas')
    expect( self.transport._read_lock.release )

    self.transport.buffer('datas')

  def test_buffer_when_raises_exception(self):
    self.transport._read_lock = mock()
    expect( self.transport._read_lock.acquire )
    with expect( mock(gevent_transport,'super') ).args(is_arg(GeventTransport), GeventTransport).returns(mock()) as parent:
      expect( parent.buffer ).args('datas').raises(Exception('fail'))
    expect( self.transport._read_lock.release )

    assert_raises(Exception, self.transport.buffer, 'datas')

  def test_write(self):
    self.transport._write_lock = mock()
    expect( self.transport._write_lock.acquire )
    with expect( mock(gevent_transport,'super') ).args(is_arg(GeventTransport), GeventTransport).returns(mock()) as parent:
      expect( parent.write ).args('datas')
    expect( self.transport._write_lock.release )

    self.transport.write('datas')

  def test_write_when_raises_an_exception(self):
    self.transport._write_lock = mock()
    expect( self.transport._write_lock.acquire )
    with expect( mock(gevent_transport,'super') ).args(is_arg(GeventTransport), GeventTransport).returns(mock()) as parent:
      expect( parent.write ).args('datas').raises(Exception('fail'))
    expect( self.transport._write_lock.release )

    assert_raises(Exception, self.transport.write, 'datas')

class GeventPoolTransportTest(Chai):

  def setUp(self):
    super(GeventPoolTransportTest,self).setUp()

    self.connection = mock()
    self.transport = GeventPoolTransport(self.connection)
    self.transport._host = 'server:1234'

  def test_init(self):
    assert_equals( bytearray(), self.transport._buffer )
    assert_true( isinstance(self.transport._read_lock,Semaphore) )
    assert_true( isinstance(self.transport.pool, Pool) )

    trans = GeventPoolTransport(self.connection, pool='inground')
    assert_equals( 'inground', trans.pool )

  def test_process_channels(self):
    chs = [ mock(), mock() ]
    self.transport._pool = mock()

    expect( self.transport._pool.spawn ).args( chs[0].process_frames )
    expect( self.transport._pool.spawn ).args( chs[1].process_frames )

    self.transport.process_channels( chs )



'''
Copyright (c) 2011-2014, Agora Games, LLC All rights reserved.

https://github.com/agoragames/haigha/blob/master/LICENSE.txt
'''

from haigha.writer import Writer
from haigha.frames import MethodFrame
from haigha.classes import ProtocolClass

from collections import deque

class QueueClass(ProtocolClass):
  '''
  Implements the AMQP Queue class
  '''

  def __init__(self, *args, **kwargs):
    super(QueueClass, self).__init__(*args, **kwargs)
    self.dispatch_map = {
      11 : self._recv_declare_ok,
      21 : self._recv_bind_ok,
      31 : self._recv_purge_ok,
      41 : self._recv_delete_ok,
      51 : self._recv_unbind_ok,
    }

    self._declare_cb = deque()
    self._bind_cb = deque()
    self._unbind_cb = deque()
    self._delete_cb = deque()
    self._purge_cb = deque()

  @property
  def name(self):
    return 'queue'

  def _cleanup(self):
    '''
    Cleanup all the local data.
    '''
    self._declare_cb = None
    self._bind_cb = None
    self._unbind_cb = None
    self._delete_cb = None
    self._purge_cb = None
    super(QueueClass,self)._cleanup()

  def declare(self, queue='', passive=False, durable=False,
      exclusive=False, auto_delete=True, nowait=True,
      arguments={}, ticket=None, cb=None):
    '''
    Declare a queue.  By default is asynchronoous but will be synchronous if nowait=False
    or a callback is defined. In synchronous mode, returns (message_count, consumer_count)

    queue - The name of the queue
    cb - An optional method which will be called with (queue_name, msg_count, consumer_count)
         if nowait=False
    '''
    nowait = nowait and self.allow_nowait() and not cb

    args = Writer()
    args.write_short(ticket or self.default_ticket).\
      write_shortstr(queue).\
      write_bits(passive, durable, exclusive, auto_delete, nowait).\
      write_table(arguments)
    self.send_frame( MethodFrame(self.channel_id, 50, 10, args) )

    if not nowait:
      self._declare_cb.append( cb )
      return self.channel.add_synchronous_cb( self._recv_declare_ok )

  def _recv_declare_ok(self, method_frame):
    queue = method_frame.args.read_shortstr()
    message_count = method_frame.args.read_long()
    consumer_count = method_frame.args.read_long()

    cb = self._declare_cb.popleft()
    if cb:
      cb( queue, message_count, consumer_count )
    return queue, message_count, consumer_count

  def bind(self, queue, exchange, routing_key='', nowait=True, arguments={},
      ticket=None, cb=None):
    '''
    bind to a queue.
    '''
    nowait = nowait and self.allow_nowait() and not cb

    args = Writer()
    args.write_short(ticket or self.default_ticket).\
      write_shortstr(queue).\
      write_shortstr(exchange).\
      write_shortstr(routing_key).\
      write_bit(nowait).\
      write_table(arguments)
    self.send_frame( MethodFrame(self.channel_id, 50, 20, args) )

    if not nowait:
      self._bind_cb.append( cb )
      self.channel.add_synchronous_cb( self._recv_bind_ok )

  def _recv_bind_ok(self, _method_frame):
    # No arguments defined.
    cb = self._bind_cb.popleft()
    if cb: cb()

  def unbind(self, queue, exchange, routing_key='', arguments={}, ticket=None, cb=None):
    '''
    Unbind a queue from an exchange.  This is always synchronous.
    '''
    args = Writer()
    args.write_short(ticket or self.default_ticket).\
      write_shortstr(queue).\
      write_shortstr(exchange).\
      write_shortstr(routing_key).\
      write_table(arguments)
    self.send_frame( MethodFrame(self.channel_id, 50, 50, args) )

    self._unbind_cb.append( cb )
    self.channel.add_synchronous_cb( self._recv_unbind_ok )

  def _recv_unbind_ok(self, _method_frame):
    # No arguments defined
    cb = self._unbind_cb.popleft()
    if cb: cb()

  def purge(self, queue, nowait=True, ticket=None, cb=None):
    '''
    Purge all messages in a queue.
    '''
    nowait = nowait and self.allow_nowait() and not cb

    args = Writer()
    args.write_short(ticket or self.default_ticket).\
      write_shortstr(queue).\
      write_bit(nowait)
    self.send_frame( MethodFrame(self.channel_id, 50, 30, args) )

    if not nowait:
      self._purge_cb.append( cb )
      return self.channel.add_synchronous_cb( self._recv_purge_ok )

  def _recv_purge_ok(self, method_frame):
    message_count = method_frame.args.read_long()
    cb = self._purge_cb.popleft()
    if cb: cb( message_count )
    return message_count

  def delete(self, queue, if_unused=False, if_empty=False, nowait=True,
    ticket=None, cb=None):
    '''
    queue delete.
    '''
    nowait = nowait and self.allow_nowait() and not cb

    args = Writer()
    args.write_short(ticket or self.default_ticket).\
      write_shortstr(queue).\
      write_bits(if_unused, if_empty, nowait)
    self.send_frame( MethodFrame(self.channel_id, 50, 40, args) )

    if not nowait:
      self._delete_cb.append( cb )
      return self.channel.add_synchronous_cb( self._recv_delete_ok )

  def _recv_delete_ok(self, method_frame):
    message_count = method_frame.args.read_long()
    cb = self._delete_cb.popleft()
    if cb: cb( message_count )
    return message_count



'''
Copyright (c) 2011-2014, Agora Games, LLC All rights reserved.

https://github.com/agoragames/haigha/blob/master/LICENSE.txt
'''

from haigha.channel import Channel
from haigha.frames import *
from haigha.classes import *
from haigha.writer import Writer
from haigha.reader import Reader
from haigha.transports import Transport
from exceptions import *

import socket
import struct
import haigha
import time

from io import BytesIO
from logging import DEBUG, INFO, WARNING, ERROR, CRITICAL
from logging import root as root_logger

# From the rabbitmq mailing list
# AMQP1180 = 0-8
# AMQP1109 = 0-9
# AMQP0091 = 0-9-1
# http://lists.rabbitmq.com/pipermail/rabbitmq-discuss/2010-July/008231.html
#PROTOCOL_HEADER = 'AMQP\x01\x01\x09\x01'
PROTOCOL_HEADER = 'AMQP\x00\x00\x09\x01'

#
# Client property info that gets sent to the server on connection startup
#
LIBRARY_PROPERTIES = {
  'library': 'Haigha',
  'library_version': haigha.__version__,
}

class Connection(object):

  class TooManyChannels(ConnectionError): '''This connection has too many channels open.  Non-fatal.'''
  class InvalidChannel(ConnectionError): '''The channel id does not correspond to an existing channel.  Non-fatal.'''

  def __init__(self, **kwargs):
    '''
    Initialize the connection.
    '''
    self._debug = kwargs.get('debug', False)
    self._logger = kwargs.get('logger', root_logger)

    self._user = kwargs.get('user', 'guest')
    self._password = kwargs.get('password', 'guest')
    self._host = kwargs.get('host', 'localhost')
    self._port = kwargs.get('port', 5672)
    self._vhost = kwargs.get('vhost', '/')

    self._connect_timeout = kwargs.get('connect_timeout', 5)
    self._sock_opts = kwargs.get('sock_opts')
    self._sock = None
    self._heartbeat = kwargs.get('heartbeat')
    self._open_cb = kwargs.get('open_cb')
    self._close_cb = kwargs.get('close_cb')

    self._login_method = kwargs.get('login_method', 'AMQPLAIN')
    self._locale = kwargs.get('locale', 'en_US')
    self._client_properties = kwargs.get('client_properties')

    self._properties = LIBRARY_PROPERTIES.copy()
    if self._client_properties:
      self._properties.update( self._client_properties )

    self._closed = False
    self._connected = False
    self._close_info = {
      'reply_code'    : 0,
      'reply_text'    : 'first connect',
      'class_id'      : 0,
      'method_id'     : 0
    }

    # Not sure what's better here, setdefaults or require the caller to pass
    # the whole thing in.
    self._class_map = kwargs.get('class_map', {}).copy()
    self._class_map.setdefault(20, ChannelClass)
    self._class_map.setdefault(40, ExchangeClass)
    self._class_map.setdefault(50, QueueClass)
    self._class_map.setdefault(60, BasicClass)
    self._class_map.setdefault(90, TransactionClass)

    self._channels = {
      0 : ConnectionChannel(self, 0, {})
    }

    # Login response seems a total hack of protocol
    # Skip the length at the beginning
    login_response = Writer()
    login_response.write_table({'LOGIN': self._user, 'PASSWORD': self._password})
    self._login_response = login_response.buffer()[4:]

    self._channel_counter = 0
    self._channel_max = 65535
    self._frame_max = 65535

    self._frames_read = 0
    self._frames_written = 0

    # Default to the socket strategy
    transport = kwargs.get('transport', 'socket')
    if not isinstance(transport, Transport):
      if transport=='event':
        from haigha.transports.event_transport import EventTransport
        self._transport = EventTransport( self )
      elif transport=='gevent':
        from haigha.transports.gevent_transport import GeventTransport
        self._transport = GeventTransport( self )
      elif transport=='gevent_pool':
        from haigha.transports.gevent_transport import GeventPoolTransport
        self._transport = GeventPoolTransport( self )
      elif transport=='socket':
        from haigha.transports.socket_transport import SocketTransport
        self._transport = SocketTransport( self )
    else:
      self._transport = transport

    self._output_frame_buffer = []
    self.connect( self._host, self._port )

  @property
  def logger(self):
    return self._logger

  @property
  def debug(self):
    return self._debug

  @property
  def frame_max(self):
    return self._frame_max

  @property
  def channel_max(self):
    return self._channel_max

  @property
  def frames_read(self):
    '''Number of frames read in the lifetime of this connection.'''
    return self._frames_read

  @property
  def frames_written(self):
    '''Number of frames written in the lifetime of this connection.'''
    return self._frames_written

  @property
  def close_info(self):
    '''Return dict with information on why this connection is closed.  Will
    return None if the connections is open.'''
    return self._close_info if (self._closed or not self._connected) else None

  @property
  def transport(self):
    '''Get the value of the current transport.'''
    return self._transport

  @property
  def synchronous(self):
    '''True if transport is synchronous, False otherwise.'''
    if self._transport==None:
      if self._close_info and len(self._close_info['reply_text'])>0:
        raise ConnectionClosed("connection is closed: %s : %s"%\
          (self._close_info['reply_code'],self._close_info['reply_text']) )
      raise ConnectionClosed("connection is closed")
    return self.transport.synchronous

  def connect(self, host, port):
    '''
    Connect to a host and port. Can be called directly, or is called by the
    strategy as it tries to find and connect to hosts.
    '''
    # Clear the connect state immediately since we're no longer connected
    # at this point.
    self._connected = False

    # Only after the socket has connected do we clear this state; closed must
    # be False so that writes can be buffered in writePacket().  The closed
    # state might have been set to True due to a socket error or a redirect.
    self._host = "%s:%d"%(host,port)
    self._closed = False
    self._close_info = {
      'reply_code'    : 0,
      'reply_text'    : 'failed to connect to %s'%(self._host),
      'class_id'      : 0,
      'method_id'     : 0
    }

    self._transport.connect( (host,port) )
    self._transport.write( PROTOCOL_HEADER )

    while self.synchronous and not self._connected:
      self.read_frames()

  def disconnect(self):
    '''
    Disconnect from the current host, but do not update the closed state. After
    the transport is disconnected, the closed state will be True if this is
    called after a protocol shutdown, or False if the disconnect was in error.

    TODO: do we really need closed vs. connected states? this only adds
    complication and the whole reconnect process has been scrapped anyway.

    '''
    self._connected = False
    if self._transport!=None:
      try:
        self._transport.disconnect()
      except Exception:
        self.logger.error("Failed to disconnect from %s", self._host, exc_info=True)
        raise
      finally:
        self._transport = None

  ###
  ### Transport methods
  ###
  def transport_closed(self, **kwargs):
    """
    Called by Transports when they close unexpectedly, not as a result of
    Connection.disconnect().

    TODO: document args
    """
    msg = 'transport to %s closed : unknown cause'%(self._host)
    self.logger.warning( kwargs.get('msg', msg) )
    self._close_info = {
      'reply_code'    : kwargs.get('reply_code',0),
      'reply_text'    : kwargs.get('msg', msg),
      'class_id'      : kwargs.get('class_id',0),
      'method_id'     : kwargs.get('method_id',0)
    }

    # We're not connected any more, but we're not closed without an explicit
    # close call. This allows the strategy to distinguish the state.
    # TODO: is this all BS?
    self._connected = False
    self._transport = None

    # Call back to a user-provided close function
    self._callback_close()

  ###
  ### Connection methods
  ###
  def _next_channel_id(self):
    '''Return the next possible channel id.  Is a circular enumeration.'''
    self._channel_counter += 1
    if self._channel_counter >= self._channel_max:
      self._channel_counter = 1
    return self._channel_counter

  def channel(self, channel_id=None):
    """
    Fetch a Channel object identified by the numeric channel_id, or
    create that object if it doesn't already exist.  If channel_id is not
    None but no channel exists for that id, will raise InvalidChannel.  If
    there are already too many channels open, will raise TooManyChannels.
    """
    if channel_id is None:
      # adjust for channel 0
      if len(self._channels)-1 >= self._channel_max:
        raise Connection.TooManyChannels( "%d channels already open, max %d",
          len(self._channels)-1, self._channel_max )
      channel_id = self._next_channel_id()
      while channel_id in self._channels:
        channel_id = self._next_channel_id()
    elif channel_id in self._channels:
      return self._channels[channel_id]
    else:
      raise Connection.InvalidChannel("%s is not a valid channel id", channel_id )

    # Call open() here so that ConnectionChannel doesn't have it called.  Could
    # also solve this other ways, but it's a HACK regardless.
    rval = Channel(self, channel_id, self._class_map)
    self._channels[ channel_id ] = rval
    rval.add_close_listener( self._channel_closed )
    rval.open()
    return rval

  def _channel_closed(self, channel):
    '''
    Close listener on a channel.
    '''
    try:
      del self._channels[ channel.channel_id ]
    except KeyError:
      pass

  def close(self, reply_code=0, reply_text='', class_id=0, method_id=0):
    '''
    Close this connection.
    '''
    self._close_info = {
      'reply_code'    : reply_code,
      'reply_text'    : reply_text,
      'class_id'      : class_id,
      'method_id'     : method_id
    }
    self._channels[0].close()

  def _callback_open(self):
    '''
    Callback to any open handler that was provided in the ctor. Handler is
    responsible for exceptions.
    '''
    if self._open_cb:
      self._open_cb()

  def _callback_close(self):
    '''
    Callback to any close handler that was provided in the ctor. Handler is
    responsible for exceptions.
    '''
    if self._close_cb:
      self._close_cb()

  def read_frames(self):
    '''
    Read frames from the transport and process them. Some transports may choose
    to do this in the background, in several threads, and so on.
    '''
    # It's possible in a concurrent environment that our transport handle has
    # gone away, so handle that cleanly.
    # TODO: Consider moving this block into Translator base class. In many
    # ways it belongs there. One of the problems though is that this is
    # essentially the read loop. Each Transport has different rules for how to
    # kick this off, and in the case of gevent, this is how a blocking call to
    # read from the socket is kicked off.
    if self._transport is None:
      return

    # Send a heartbeat (if needed)
    self._channels[0].send_heartbeat()

    data = self._transport.read( self._heartbeat )
    if data is None:
      return

    reader = Reader( data )
    p_channels = set()

    for frame in Frame.read_frames( reader ):
      if self._debug > 1:
        self.logger.debug( "READ: %s", frame )
      self._frames_read += 1
      ch = self.channel( frame.channel_id )
      ch.buffer_frame( frame )
      p_channels.add( ch )

    self._transport.process_channels( p_channels )

    # HACK: read the buffer contents and re-buffer.  Would prefer to pass
    # buffer back, but there's no good way of asking the total size of the
    # buffer, comparing to tell(), and then re-buffering.  There's also no
    # ability to clear the buffer up to the current position. It would be
    # awesome if we could free that memory without a new allocation.
    if reader.tell() < len(data):
      self._transport.buffer( data[reader.tell():] )

  def _flush_buffered_frames(self):
    '''
    Callback when protocol has been initialized on channel 0 and we're ready
    to send out frames to set up any channels that have been created.
    '''
    # In the rare case (a bug) where this is called but send_frame thinks
    # they should be buffered, don't clobber.
    frames = self._output_frame_buffer
    self._output_frame_buffer = []
    for frame in frames:
      self.send_frame( frame )

  def send_frame(self, frame):
    '''
    Send a single frame. If there is no transport or we're not connected yet,
    append to the output buffer, else send immediately to the socket. This is
    called from within the MethodFrames.
    '''
    if self._closed:
      if self._close_info and len(self._close_info['reply_text'])>0:
        raise ConnectionClosed("connection is closed: %s : %s"%\
          (self._close_info['reply_code'],self._close_info['reply_text']) )
      raise ConnectionClosed("connection is closed")

    if self._transport==None or (not self._connected and frame.channel_id!=0):
      self._output_frame_buffer.append( frame )
      return

    if self._debug > 1:
      self.logger.debug( "WRITE: %s", frame )

    buf = bytearray()
    frame.write_frame(buf)
    self._transport.write( buf )

    self._frames_written += 1


class ConnectionChannel(Channel):
  '''
  A special channel for the Connection class.  It's used for performing the special
  methods only available on the main connection channel.  It's also partly used to
  hide the 'connection' protocol implementation, which would show up as a property,
  from the more useful 'connection' property that is a handle to a Channel's
  Connection object.
  '''

  def __init__(self, *args):
    super(ConnectionChannel,self).__init__(*args)

    self._method_map = {
      10 : self._recv_start,
      20 : self._recv_secure,
      30 : self._recv_tune,
      41 : self._recv_open_ok,
      50 : self._recv_close,
      51 : self._recv_close_ok,
    }
    self._last_heartbeat_send = 0

  def dispatch(self, frame):
    '''
    Override the default dispatch since we don't need the rest of the stack.
    '''
    if frame.type()==HeartbeatFrame.type():
      self.send_heartbeat()

    elif frame.type()==MethodFrame.type():
      if frame.class_id==10:
        cb = self._method_map.get( frame.method_id )
        if cb:
          cb( frame )
        else:
          raise Channel.InvalidMethod("unsupported method %d on channel %d",
            frame.method_id, self.channel_id )
      else:
        raise Channel.InvalidClass( "class %d is not supported on channel %d",
          frame.class_id, self.channel_id )

    else:
      raise Frame.InvalidFrameType("frame type %d is not supported on channel %d",
        frame.type(), self.channel_id)

  def close(self, reply_code=0, reply_text='', class_id=0, method_id=0):
    '''
    Close the main connection connection channel.
    '''
    self._send_close()

  def send_heartbeat(self):
    '''
    Send a heartbeat if needed. Tracks last heartbeat send time.
    '''
    # Note that this does not take into account the time that we last sent a
    # frame. Hearbeats are so small the effect should be quite limited. Also
    # note that we're looking for something near to our scheduled interval,
    # because if this is exact, then we'll likely actually send a heartbeat
    # at twice the period, which could cause a broker to kill the connection
    # if the period is large enough. The 90% bound was chosen arbitrarily but
    # seems a sensible enough default
    if self.connection._heartbeat:
      if time.time() >= (self._last_heartbeat_send + 0.9*self.connection._heartbeat):
        self.send_frame( HeartbeatFrame(self.channel_id) )
        self._last_heartbeat_send = time.time()

  def _recv_start(self, method_frame):
    self.connection._closed = False
    self._send_start_ok()

  def _send_start_ok(self):
    '''Send the start_ok message.'''
    args = Writer()
    args.write_table(self.connection._properties)
    args.write_shortstr(self.connection._login_method)
    args.write_longstr(self.connection._login_response)
    args.write_shortstr(self.connection._locale)
    self.send_frame( MethodFrame(self.channel_id, 10, 11, args) )

  def _recv_tune(self, method_frame):
    self.connection._channel_max = method_frame.args.read_short() or self.connection._channel_max
    self.connection._frame_max = method_frame.args.read_long() or self.connection._frame_max

    # Note that 'is' test is required here, as 0 and None are distinct
    if self.connection._heartbeat is None:
      self.connection._heartbeat = method_frame.args.read_short()

    self._send_tune_ok()
    self._send_open()

    # 4.2.7: The client should start sending heartbeats after receiving a
    # Connection.Tune method
    self.send_heartbeat()

  def _send_tune_ok(self):
    args = Writer()
    args.write_short( self.connection._channel_max )
    args.write_long( self.connection._frame_max )

    if self.connection._heartbeat:
      args.write_short( self.connection._heartbeat )
    else:
      args.write_short( 0 )

    self.send_frame( MethodFrame(self.channel_id, 10, 31, args) )

  def _recv_secure(self, method_frame):
    self._send_open()

  def _send_open(self):
    args = Writer()
    args.write_shortstr(self.connection._vhost)
    args.write_shortstr('')
    args.write_bit(True)  # insist flag for older amqp, not used in 0.9.1

    self.send_frame( MethodFrame(self.channel_id, 10, 40, args) )

  def _recv_open_ok(self, method_frame):
    self.connection._connected = True
    self.connection._flush_buffered_frames()
    self.connection._callback_open()

  def _send_close(self):
    args = Writer()
    args.write_short( self.connection._close_info['reply_code'] )
    args.write_shortstr( self.connection._close_info['reply_text'][:255] )
    args.write_short( self.connection._close_info['class_id'] )
    args.write_short( self.connection._close_info['method_id'] )
    self.send_frame( MethodFrame(self.channel_id, 10, 50, args) )

  def _recv_close(self, method_frame):
    self.connection._close_info = {
      'reply_code'    : method_frame.args.read_short(),
      'reply_text'    : method_frame.args.read_shortstr(),
      'class_id'      : method_frame.args.read_short(),
      'method_id'     : method_frame.args.read_short()
    }

    # TODO: wait to disconnect until the close_ok has been flushed, but that's
    # a pain
    self._send_close_ok()

    self.connection._closed = True
    self.connection.disconnect()
    self.connection._callback_close()

  def _send_close_ok(self):
    self.send_frame( MethodFrame(self.channel_id, 10, 51) )

  def _recv_close_ok(self, method_frame):
    self.connection._closed = True
    self.connection.disconnect()
    self.connection._callback_close()



'''
Copyright (c) 2011-2014, Agora Games, LLC All rights reserved.

https://github.com/agoragames/haigha/blob/master/LICENSE.txt
'''

import logging
from chai import Chai

from haigha import connection, __version__
from haigha.connection import Connection, ConnectionChannel
from haigha.channel import Channel
from haigha.frames import *
from haigha.classes import *

from haigha.transports import event_transport
from haigha.transports import gevent_transport
from haigha.transports import socket_transport

class ConnectionTest(Chai):

  def setUp(self):
    super(ConnectionTest,self).setUp()

    self.connection = Connection.__new__( Connection )
    self.connection._debug = False
    self.connection._logger = self.mock()
    self.connection._user = 'guest'
    self.connection._password = 'guest'
    self.connection._host = 'localhost'
    self.connection._vhost = '/'
    self.connection._connect_timeout = 5
    self.connection._sock_opts = None
    self.connection._sock = None # mock anything?
    self.connection._heartbeat = None
    self.connection._open_cb = self.mock()
    self.connection._close_cb = self.mock()
    self.connection._login_method = 'AMQPLAIN'
    self.connection._locale = 'en_US'
    self.connection._client_properties = None
    self.connection._properties = {
      'library': 'Haigha',
      'library_version': 'x.y.z',
    }
    self.connection._closed = False
    self.connection._connected = False
    self.connection._close_info = {
      'reply_code'    : 0,
      'reply_text'    : 'first connect',
      'class_id'      : 0,
      'method_id'     : 0
    }
    self.connection._class_map = {}
    self.connection._channels = {
      0 : self.mock()
    }
    self.connection._login_response = 'loginresponse'
    self.connection._channel_counter = 0
    self.connection._channel_max = 65535
    self.connection._frame_max = 65535
    self.connection._frames_read = 0
    self.connection._frames_written = 0
    self.connection._strategy = self.mock()
    self.connection._output_frame_buffer = []
    self.connection._transport = mock()

  def test_init_without_keyword_args(self):
    conn = Connection.__new__( Connection )
    strategy = mock()
    mock( connection, 'ConnectionChannel' )

    expect(connection.ConnectionChannel).args( conn, 0, {} ).returns( 'connection_channel' )
    expect(socket_transport.SocketTransport).args( conn ).returns( 'SocketTransport' )
    expect(conn.connect).args( 'localhost', 5672 )

    conn.__init__()

    self.assertFalse( conn._debug )
    self.assertEqual( logging.root, conn._logger )
    self.assertEqual( 'guest', conn._user )
    self.assertEqual( 'guest', conn._password )
    self.assertEqual( 'localhost', conn._host )
    self.assertEqual( 5672, conn._port )
    self.assertEqual( '/', conn._vhost )
    self.assertEqual( 5, conn._connect_timeout )
    self.assertEqual( None, conn._sock_opts )
    self.assertEqual( None, conn._sock )
    self.assertEqual( None, conn._heartbeat )
    self.assertEqual( None, conn._open_cb )
    self.assertEqual( None, conn._close_cb )
    self.assertEqual( 'AMQPLAIN', conn._login_method )
    self.assertEqual( 'en_US', conn._locale )
    self.assertEqual( None, conn._client_properties )
    self.assertEqual( conn._properties, {
      'library': 'Haigha',
      'library_version': __version__,
    } )
    self.assertFalse( conn._closed )
    self.assertFalse( conn._connected )
    self.assertEqual( conn._close_info, {
      'reply_code'    : 0,
      'reply_text'    : 'first connect',
      'class_id'      : 0,
      'method_id'     : 0
    } )
    assert_equals( {
      20: ChannelClass,
      40: ExchangeClass,
      50: QueueClass,
      60: BasicClass,
      90: TransactionClass
    }, conn._class_map )
    self.assertEqual( {0:'connection_channel'}, conn._channels )
    self.assertEqual( '\x05LOGINS\x00\x00\x00\x05guest\x08PASSWORDS\x00\x00\x00\x05guest', conn._login_response )
    self.assertEqual( 0, conn._channel_counter )
    self.assertEqual( 65535, conn._channel_max )
    self.assertEqual( 65535, conn._frame_max )
    self.assertEqual( [], conn._output_frame_buffer )
    self.assertEqual( 'SocketTransport', conn._transport )

  def test_init_with_event_transport(self):
    conn = Connection.__new__( Connection )
    strategy = mock()
    mock( connection, 'ConnectionChannel' )

    expect(connection.ConnectionChannel).args( conn, 0, {} ).returns( 'connection_channel' )
    expect(event_transport.EventTransport).args( conn ).returns( 'EventTransport' )
    expect(conn.connect).args( 'localhost', 5672 )

    conn.__init__(transport='event')

  def test_properties(self):
    self.assertEqual( self.connection._logger, self.connection.logger )
    self.assertEqual( self.connection._debug, self.connection.debug )
    self.assertEqual( self.connection._frame_max, self.connection.frame_max )
    self.assertEqual( self.connection._channel_max, self.connection.channel_max )
    self.assertEqual( self.connection._frames_read, self.connection.frames_read )
    self.assertEqual( self.connection._frames_written, self.connection.frames_written )

  def test_synchronous_when_no_transport(self):
    self.connection._transport = None
    with assert_raises( connection.ConnectionClosed ):
      self.connection.synchronous

    self.connection._close_info = {
      'reply_code':100,
      'reply_text':'breakdown'
    }
    with assert_raises( connection.ConnectionClosed ):
      self.connection.synchronous

  def test_synchronous_when_transport(self):
    self.connection._transport.synchronous = True
    assert_true( self.connection.synchronous )
    self.connection._transport.synchronous = False
    assert_false( self.connection.synchronous )

  def test_connect_when_asynchronous_transport(self):
    self.connection._transport.synchronous = False
    self.connection._connected = 'maybe'
    self.connection._closed = 'possibly'
    self.connection._debug = 'sure'
    self.connection._connect_timeout = 42
    self.connection._sock_opts = {
      ('f1','t1') : 5,
      ('f2','t2') : 6
    }

    expect( self.connection._transport.connect ).args( ('host',5672) )
    expect( self.connection._transport.write ).args( 'AMQP\x00\x00\x09\x01' )

    self.connection.connect( 'host', 5672 )
    assert_false( self.connection._connected )
    assert_false( self.connection._closed )
    assert_equals( self.connection._close_info,
      {
      'reply_code'    : 0,
      'reply_text'    : 'failed to connect to host:5672',
      'class_id'      : 0,
      'method_id'     : 0
      } )
    assert_equals( 'host:5672', self.connection._host )

  def test_connect_when_synchronous_transport(self):
    self.connection._transport.synchronous = True
    self.connection._connected = 'maybe'
    self.connection._closed = 'possibly'
    self.connection._debug = 'sure'
    self.connection._connect_timeout = 42
    self.connection._sock_opts = {
      ('f1','t1') : 5,
      ('f2','t2') : 6
    }

    expect( self.connection._transport.connect ).args( ('host',5672) )
    expect( self.connection._transport.write ).args( 'AMQP\x00\x00\x09\x01' )

    expect( self.connection.read_frames )
    expect( self.connection.read_frames ).side_effect(
      lambda: setattr(self.connection,'_connected',True) )

    self.connection.connect( 'host', 5672 )
    assert_true( self.connection._connected )
    assert_false( self.connection._closed )
    assert_equals( self.connection._close_info,
      {
      'reply_code'    : 0,
      'reply_text'    : 'failed to connect to host:5672',
      'class_id'      : 0,
      'method_id'     : 0
      } )
    assert_equals( 'host:5672', self.connection._host )

  def test_disconnect_when_transport_disconnects(self):
    self.connection._connected = 'yup'

    expect( self.connection._transport.disconnect )
    self.connection.disconnect()

    assert_false( self.connection._connected )
    assert_equals( None, self.connection._transport )

  def test_disconnect_when_transport_disconnects_with_error(self):
    self.connection._connected = 'yup'
    self.connection._host = 'server'

    expect( self.connection._transport.disconnect ).raises( RuntimeError('fail') )
    expect( self.connection.logger.error ).args( "Failed to disconnect from %s", 'server', exc_info=True )
    assert_raises( RuntimeError, self.connection.disconnect )

    assert_false( self.connection._connected )
    assert_equals( None, self.connection._transport )

  def test_disconnect_when_systemexit(self):
    self.connection._connected = 'yup'
    self.connection._host = 'server'

    expect( self.connection._transport.disconnect ).raises( SystemExit() )
    stub( self.connection.logger.error )
    assert_raises( SystemExit, self.connection.disconnect )

    assert_false( self.connection._connected )
    assert_equals( None, self.connection._transport )


  def test_transport_closed_with_no_args(self):
    self.connection._host = 'server'
    self.connection._connected = 'yes'

    expect( self.connection.logger.warning ).args( 'transport to server closed : unknown cause' )
    expect( self.connection._callback_close )

    self.connection.transport_closed()

    assert_equals( 0, self.connection._close_info['reply_code'] )
    assert_equals( 'transport to server closed : unknown cause', self.connection._close_info['reply_text'] )
    assert_equals( 0, self.connection._close_info['class_id'] )
    assert_equals( 0, self.connection._close_info['method_id'] )

  def test_next_channel_id_when_less_than_max(self):
    self.connection._channel_counter = 32
    self.connection._channel_max = 23423
    assert_equals( 33, self.connection._next_channel_id() )

  def test_next_channel_id_when_at_max(self):
    self.connection._channel_counter = 32
    self.connection._channel_max = 32
    assert_equals( 1, self.connection._next_channel_id() )

  def test_channel_creates_new_when_not_at_limit(self):
    ch = mock()
    expect( self.connection._next_channel_id ).returns( 1 )
    mock( connection, 'Channel' )
    expect( connection.Channel ).args(
      self.connection, 1, self.connection._class_map).returns( ch )
    expect( ch.add_close_listener ).args( self.connection._channel_closed )
    expect( ch.open )

    self.assert_equals( ch, self.connection.channel() )
    self.assert_equals( ch, self.connection._channels[1] )

  def test_channel_finds_the_first_free_channel_id(self):
    self.connection._channels[1] = 'foo'
    self.connection._channels[2] = 'bar'
    self.connection._channels[4] = 'cat'
    ch = mock()
    expect( self.connection._next_channel_id ).returns( 1 )
    expect( self.connection._next_channel_id ).returns( 2 )
    expect( self.connection._next_channel_id ).returns( 3 )
    mock( connection, 'Channel' )
    expect( connection.Channel ).args(
      self.connection, 3, self.connection._class_map ).returns( ch )
    expect( ch.add_close_listener ).args( self.connection._channel_closed )
    expect( ch.open )

    self.assert_equals( ch, self.connection.channel() )
    self.assert_equals( ch, self.connection._channels[3] )

  def test_channel_raises_toomanychannels(self):
    self.connection._channels[1] = 'foo'
    self.connection._channels[2] = 'bar'
    self.connection._channels[4] = 'cat'
    self.connection._channel_max = 3
    assert_raises( Connection.TooManyChannels, self.connection.channel )

  def test_channel_returns_cached_instance_if_known(self):
    self.connection._channels[1] = 'foo'
    assert_equals( 'foo', self.connection.channel(1) )

  def test_channel_raises_invalidchannel_if_unknown_id(self):
    assert_raises( Connection.InvalidChannel, self.connection.channel, 42 )

  def test_channel_closed(self):
    ch = mock()
    ch.channel_id = 42
    self.connection._channels[42] = ch

    self.connection._channel_closed(ch)
    assert_false( 42 in self.connection._channels )

    ch.channel_id = 500424834
    self.connection._channel_closed(ch)

  def test_close(self):
    self.connection._channels[0] = mock()
    expect( self.connection._channels[0].close ).times(2)

    self.connection.close()
    assert_equals( {'reply_code':0, 'reply_text':'', 'class_id':0, 'method_id':0},
      self.connection._close_info )

    self.connection.close(1, 'foo', 2, 3)
    assert_equals( {'reply_code':1, 'reply_text':'foo', 'class_id':2, 'method_id':3},
      self.connection._close_info )

  def test_callback_open_when_no_cb(self):
    self.connection._open_cb = None
    self.connection._callback_open()

  def test_callback_open_when_user_cb(self):
    self.connection._open_cb = mock()
    expect( self.connection._open_cb )
    self.connection._callback_open()

  def test_callback_open_raises_when_user_cb_does(self):
    self.connection._open_cb = mock()
    expect( self.connection._open_cb ).raises( SystemExit() )
    assert_raises( SystemExit, self.connection._callback_open )

  def test_callback_close_when_no_cb(self):
    self.connection._close_cb = None
    self.connection._callback_close()

  def test_callback_close_when_user_cb(self):
    self.connection._close_cb = mock()
    expect( self.connection._close_cb )
    self.connection._callback_close()

  def test_callback_close_raises_when_user_cb_does(self):
    self.connection._close_cb = mock()
    expect( self.connection._close_cb ).raises( SystemExit() )
    assert_raises( SystemExit, self.connection._callback_close )

  def test_read_frames_when_no_transport(self):
    self.connection._transport = None
    self.connection.read_frames()
    assert_equals( 0, self.connection._frames_read )

  def test_read_frames_when_transport_returns_no_data(self):
    self.connection._heartbeat = None
    expect( self.connection._channels[0].send_heartbeat )
    expect( self.connection._transport.read ).args(None).returns( None )
    self.connection.read_frames()
    assert_equals( 0, self.connection._frames_read )

  def test_read_frames_when_transport_when_frame_data_and_no_debug_and_no_buffer(self):
    reader = mock()
    frame = mock()
    frame.channel_id = 42
    channel = mock()
    mock( connection, 'Reader' )
    self.connection._heartbeat = 3

    expect( self.connection._channels[0].send_heartbeat )
    expect( self.connection._transport.read ).args(3).returns( 'data' )
    expect( connection.Reader ).args( 'data' ).returns( reader )
    expect( connection.Frame.read_frames ).args( reader ).returns( [frame] )
    expect( self.connection.channel ).args( 42 ).returns( channel )
    expect( channel.buffer_frame ).args( frame )
    expect( self.connection._transport.process_channels ).args( set([channel]) )
    expect( reader.tell ).returns( 4 )

    self.connection.read_frames()
    assert_equals( 1, self.connection._frames_read )

  def test_read_frames_when_transport_when_frame_data_and_debug_and_buffer(self):
    reader = mock()
    frame = mock()
    frame.channel_id = 42
    channel = mock()
    mock( connection, 'Reader' )
    self.connection._debug = 2

    expect( self.connection._channels[0].send_heartbeat )
    expect( self.connection._transport.read ).args(None).returns( 'data' )
    expect( connection.Reader ).args( 'data' ).returns( reader )
    expect( connection.Frame.read_frames ).args( reader ).returns( [frame] )
    expect( self.connection.logger.debug ).args( 'READ: %s', frame )
    expect( self.connection.channel ).args( 42 ).returns( channel )
    expect( channel.buffer_frame ).args( frame )
    expect( self.connection._transport.process_channels ).args( set([channel]) )
    expect( reader.tell ).times(2).returns( 2 )
    expect( self.connection._transport.buffer ).args( 'ta' )

    self.connection.read_frames()
    assert_equals( 1, self.connection._frames_read )

  def test_flush_buffered_frames(self):
    self.connection._output_frame_buffer = ['frame1', 'frame2']
    expect( self.connection.send_frame ).args( 'frame1' )
    expect( self.connection.send_frame ).args( 'frame2' )

    self.connection._flush_buffered_frames()
    assert_equals( [], self.connection._output_frame_buffer )

  def test_send_frame_when_connected_and_transport_and_no_debug(self):
    frame = mock()
    expect( frame.write_frame ).args( var('ba') )
    expect( self.connection._transport.write ).args( var('ba') )

    self.connection._connected = True
    self.connection.send_frame( frame )
    assert_true( isinstance(var('ba').value, bytearray) )
    assert_equals( 1, self.connection._frames_written )

  def test_send_frame_when_not_connected_and_not_channel_0(self):
    frame = mock()
    frame.channel_id = 42
    stub( frame.write_frame )
    stub( self.connection._transport.write )

    self.connection._connected = False
    self.connection.send_frame( frame )
    assert_equals( [frame], self.connection._output_frame_buffer )

  def test_send_frame_when_not_connected_and_channel_0(self):
    frame = mock()
    frame.channel_id = 0
    expect( frame.write_frame ).args( var('ba') )
    expect( self.connection._transport.write ).args( var('ba') )

    self.connection._connected = False
    self.connection.send_frame( frame )
    assert_true( isinstance(var('ba').value, bytearray) )
    assert_equals( 1, self.connection._frames_written )

  def test_send_frame_when_debugging(self):
    frame = mock()
    expect( self.connection.logger.debug ).args( 'WRITE: %s', frame )
    expect( frame.write_frame ).args( var('ba') )
    expect( self.connection._transport.write ).args( var('ba') )

    self.connection._connected = True
    self.connection._debug = 2
    self.connection.send_frame( frame )
    assert_true( isinstance(var('ba').value, bytearray) )
    assert_equals( 1, self.connection._frames_written )

  def test_send_frame_when_closed(self):
    self.connection._closed = True
    self.connection._close_info['reply_text'] = 'failed'
    assert_raises( connection.ConnectionClosed,
      self.connection.send_frame, 'frame' )

    self.connection._close_info['reply_text'] = ''
    assert_raises( connection.ConnectionClosed,
      self.connection.send_frame, 'frame' )

    self.connection._close_info = None
    assert_raises( connection.ConnectionClosed,
      self.connection.send_frame, 'frame' )


class ConnectionChannelTest(Chai):

  def setUp(self):
    super(ConnectionChannelTest,self).setUp()
    self.connection = mock()
    self.ch = ConnectionChannel(self.connection, 0, {})

  def test_init(self):
    mock( connection, 'super' )
    with expect( connection, 'super' ).args( is_arg(ConnectionChannel), ConnectionChannel ).returns(mock()) as s:
      expect( s.__init__ ).args( 'a', 'b' )

    c = ConnectionChannel('a','b')
    assert_equals( c._method_map ,
      {
        10 : c._recv_start,
        20 : c._recv_secure,
        30 : c._recv_tune,
        41 : c._recv_open_ok,
        50 : c._recv_close,
        51 : c._recv_close_ok,
      }
    )
    assert_equal( 0, c._last_heartbeat_send )

  def test_dispatch_on_heartbeat_frame(self):
    frame = mock()

    expect( frame.type ).at_least(1).returns( HeartbeatFrame.type() )
    expect( self.ch.send_heartbeat )

    self.ch.dispatch( frame )

  def test_dispatch_method_frame_class_10(self):
    frame = mock()
    frame.class_id = 10
    frame.method_id = 10
    self.ch._method_map[10] = mock()

    expect( frame.type ).at_least(1).returns( MethodFrame.type() )
    expect( self.ch._method_map[10] ).args( frame )

    self.ch.dispatch( frame )

  def test_dispatch_method_frame_raises_invalidmethod(self):
    frame = mock()
    frame.class_id = 10
    frame.method_id = 500

    expect( frame.type ).at_least(1).returns( MethodFrame.type() )

    with assert_raises( Channel.InvalidMethod ):
      self.ch.dispatch( frame )

  def test_dispatch_method_frame_raises_invalidclass(self):
    frame = mock()
    frame.class_id = 11
    frame.method_id = 10

    expect( frame.type ).at_least(1).returns( MethodFrame.type() )

    with assert_raises( Channel.InvalidClass ):
      self.ch.dispatch( frame )

  def test_dispatch_method_frame_raises_invalidframetype(self):
    frame = mock()

    expect( frame.type ).at_least(1).returns( HeaderFrame.type() )

    with assert_raises( Frame.InvalidFrameType ):
      self.ch.dispatch( frame )

  def test_close(self):
    expect( self.ch._send_close )
    self.ch.close()

  def test_send_heartbeat_when_no_heartbeat(self):
    stub( self.ch.send_frame )
    self.ch.connection._heartbeat = None

    self.ch.send_heartbeat()

  def test_send_heartbeat_when_not_sent_yet(self):
    mock( connection, 'time' )
    self.ch.connection._heartbeat = 3
    self.ch._last_heartbeat_send = 0

    expect( connection.time.time ).returns( 4200.3 ).times(2)
    expect( self.ch.send_frame ).args( HeartbeatFrame )

    self.ch.send_heartbeat()
    assert_equals( 4200.3, self.ch._last_heartbeat_send )

  def test_send_heartbeat_when_sent_long_ago(self):
    mock( connection, 'time' )
    self.ch.connection._heartbeat = 3
    self.ch._last_heartbeat_send = 4196

    expect( connection.time.time ).returns( 4200.3 ).times(2)
    expect( self.ch.send_frame ).args( HeartbeatFrame )

    self.ch.send_heartbeat()
    assert_equals( 4200.3, self.ch._last_heartbeat_send )

  def test_send_heart_when_sent_recently(self):
    mock( connection, 'time' )
    self.ch.connection._heartbeat = 3
    self.ch._last_heartbeat_send = 4199

    expect( connection.time.time ).returns( 4200.3 )
    stub( self.ch.send_frame )

    self.ch.send_heartbeat()
    assert_equals( 4199, self.ch._last_heartbeat_send )

  def test_recv_start(self):
    expect( self.ch._send_start_ok )
    self.ch.connection._closed = 'maybe'

    self.ch._recv_start('frame')
    assert_false( self.ch.connection._closed )

  def test_send_start_ok(self):
    self.ch.connection._properties = 'props'
    self.ch.connection._login_method = 'please'
    self.ch.connection._login_response = 'thanks'
    self.ch.connection._locale = 'home'

    with expect(mock(connection,'Writer')).returns(mock()) as writer:
      expect( writer.write_table ).args( 'props' )
      expect( writer.write_shortstr ).args( 'please' )
      expect( writer.write_longstr ).args( 'thanks' )
      expect( writer.write_shortstr ).args( 'home' )

      expect(mock(connection,'MethodFrame')).args(0,10,11,writer).returns('frame')
      expect( self.ch.send_frame ).args( 'frame' )

    self.ch._send_start_ok()

  def test_recv_tune_when_no_broker_max_and_defined_heartbeat(self):
    self.ch.connection._channel_max = 42
    self.ch.connection._frame_max = 43
    self.ch.connection._heartbeat = 8

    frame = mock()
    expect( frame.args.read_short ).returns(0)
    expect( frame.args.read_long ).returns(0)

    expect( self.ch._send_tune_ok )
    expect( self.ch._send_open )
    expect( self.ch.send_heartbeat )

    self.ch._recv_tune(frame)
    assert_equals( 42, self.ch.connection._channel_max )
    assert_equals( 43, self.ch.connection._frame_max )
    assert_equals( 8, self.ch.connection._heartbeat )

  def test_recv_tune_when_broker_max_and_undefined_heartbeat(self):
    self.ch.connection._channel_max = 42
    self.ch.connection._frame_max = 43
    self.ch.connection._heartbeat = None

    frame = mock()
    expect( frame.args.read_short ).returns(500)
    expect( frame.args.read_long ).returns(501)
    expect( frame.args.read_short ).returns(7)

    expect( self.ch._send_tune_ok )
    expect( self.ch._send_open )
    expect( self.ch.send_heartbeat )

    self.ch._recv_tune(frame)
    assert_equals( 500, self.ch.connection._channel_max )
    assert_equals( 501, self.ch.connection._frame_max )
    assert_equals( 7, self.ch.connection._heartbeat )

  def test_send_tune_ok_when_heartbeat(self):
    self.ch.connection._channel_max = 42
    self.ch.connection._frame_max = 43
    self.ch.connection._heartbeat = 8

    with expect(mock(connection,'Writer')).returns(mock()) as writer:
      expect( writer.write_short ).args( 42 )
      expect( writer.write_long ).args( 43 )
      expect( writer.write_short ).args( 8 )

      expect(mock(connection,'MethodFrame')).args(0,10,31,writer).returns('frame')
      expect( self.ch.send_frame ).args( 'frame' )

    self.ch._send_tune_ok()

  def test_send_tune_ok_when_no_heartbeat(self):
    self.ch.connection._channel_max = 42
    self.ch.connection._frame_max = 43
    self.ch.connection._heartbeat = None

    with expect(mock(connection,'Writer')).returns(mock()) as writer:
      expect( writer.write_short ).args( 42 )
      expect( writer.write_long ).args( 43 )
      expect( writer.write_short ).args( 0 )

      expect(mock(connection,'MethodFrame')).args(0,10,31,writer).returns('frame')
      expect( self.ch.send_frame ).args( 'frame' )

    self.ch._send_tune_ok()

  def test_recv_secure(self):
    expect( self.ch._send_open )
    self.ch._recv_secure('frame')

  def test_send_open(self):
    self.connection._vhost = '/foo'

    with expect(mock(connection,'Writer')).returns(mock()) as writer:
      expect( writer.write_shortstr ).args( '/foo' )
      expect( writer.write_shortstr ).args( '' )
      expect( writer.write_bit ).args( True )

      expect(mock(connection,'MethodFrame')).args(0,10,40,writer).returns('frame')
      expect( self.ch.send_frame ).args( 'frame' )

    self.ch._send_open()

  def test_recv_open_ok(self):
    self.ch.connection._connected = False
    expect( self.ch.connection._flush_buffered_frames )
    expect( self.ch.connection._callback_open )

    self.ch._recv_open_ok('frame')
    assert_true( self.ch.connection._connected )

  def test_send_close(self):
    self.ch.connection._close_info = {
      'reply_code':42,
      'reply_text':'wrong answer'*60,
      'class_id':4,
      'method_id':20,
    }

    with expect(mock(connection,'Writer')).returns(mock()) as writer:
      expect( writer.write_short ).args( 42 )
      expect( writer.write_shortstr ).args( ('wrong answer'*60)[:255] )
      expect( writer.write_short ).args( 4 )
      expect( writer.write_short ).args( 20 )

      expect(mock(connection,'MethodFrame')).args(0,10,50,writer).returns('frame')
      expect( self.ch.send_frame ).args( 'frame' )

    self.ch._send_close()

  def test_recv_close(self):
    self.ch.connection._closed = False

    frame = mock()
    expect( frame.args.read_short ).returns( 42 )
    expect( frame.args.read_shortstr ).returns( 'wrong answer' )
    expect( frame.args.read_short ).returns( 4 )
    expect( frame.args.read_short ).returns( 20 )

    expect( self.ch._send_close_ok )
    expect( self.ch.connection.disconnect )
    expect( self.ch.connection._callback_close )

    self.ch._recv_close( frame )
    assert_equals( self.ch.connection._close_info, {
      'reply_code':42,
      'reply_text':'wrong answer',
      'class_id':4,
      'method_id':20,
    })
    assert_true( self.ch.connection._closed )

  def test_send_close_ok(self):
    expect(mock(connection,'MethodFrame')).args(0,10,51).returns('frame')
    expect( self.ch.send_frame ).args( 'frame' )
    self.ch._send_close_ok()

  def test_recv_close_ok(self):
    self.ch.connection._closed = False
    expect( self.ch.connection.disconnect )
    expect( self.ch.connection._callback_close )

    self.ch._recv_close_ok('frame')
    assert_true( self.ch.connection._closed )



'''
Copyright (c) 2011-2014, Agora Games, LLC All rights reserved.

https://github.com/agoragames/haigha/blob/master/LICENSE.txt
'''

from collections import deque

from chai import Chai

from haigha.channel_pool import ChannelPool

class ChannelPoolTest(Chai):

  def test_init(self):
    c = ChannelPool('connection')
    self.assertEquals('connection', c._connection)
    self.assertEquals(set(), c._free_channels)
    assert_equals( None, c._size )
    assert_equals( 0, c._channels )
    assert_equals( deque(), c._queue )

    c = ChannelPool('connection', size=50)
    self.assertEquals('connection', c._connection)
    self.assertEquals(set(), c._free_channels)
    assert_equals( 50, c._size )
    assert_equals( 0, c._channels )
    assert_equals( deque(), c._queue )

  def test_publish_without_user_cb(self):
    ch = mock()
    cp = ChannelPool(None)

    expect(cp._get_channel).returns( ch )
    expect(ch.publish_synchronous).args( 'arg1', 'arg2', cb=var('cb'), doit='harder' )

    cp.publish( 'arg1', 'arg2', doit='harder' )
    assert_equals( set(), cp._free_channels )

    # run committed callback
    var('cb').value()
    assert_equals( set([ch]), cp._free_channels )

  def test_publish_with_user_cb(self):
    ch = mock()
    cp = ChannelPool(None)
    user_cb = mock()

    expect(cp._get_channel).returns( ch )
    expect(ch.publish_synchronous).args( 'arg1', 'arg2', cb=var('cb'), doit='harder' )

    cp.publish( 'arg1', 'arg2', cb=user_cb, doit='harder' )
    assert_equals( set(), cp._free_channels )

    expect(user_cb)
    var('cb').value()
    assert_equals( set([ch]), cp._free_channels )

  def test_publish_resends_queued_messages_if_channel_is_active(self):
    ch = mock()
    cp = ChannelPool(None)
    user_cb = mock()
    ch.active = True
    ch.closed = False
    cp._queue.append( (('a1', 'a2'), {'cb':'foo', 'yo':'dawg'}) )

    expect(cp._get_channel).returns( ch )
    expect(ch.publish_synchronous).args( 'arg1', 'arg2', cb=var('cb'), doit='harder' )

    cp.publish( 'arg1', 'arg2', cb=user_cb, doit='harder' )
    assert_equals( set(), cp._free_channels )
    assert_equals( 1, len(cp._queue) )

    expect(cp._process_queue)
    expect(user_cb)
    var('cb').value()
    assert_equals( set([ch]), cp._free_channels )

  def test_publish_does_not_resend_queued_messages_if_channel_is_inactive(self):
    ch = mock()
    cp = ChannelPool(None)
    user_cb = mock()
    ch.active = True
    ch.closed = False
    cp._queue.append( (('a1', 'a2'), {'cb':'foo', 'yo':'dawg'}) )

    expect(cp._get_channel).returns( ch )
    expect(ch.publish_synchronous).args( 'arg1', 'arg2', cb=var('cb'), doit='harder' )

    cp.publish( 'arg1', 'arg2', cb=user_cb, doit='harder' )
    assert_equals( set(), cp._free_channels )
    assert_equals( 1, len(cp._queue) )

    ch.active = False

    stub(cp._process_queue)
    expect(user_cb)
    var('cb').value()
    assert_equals( set([ch]), cp._free_channels )
    assert_equals( 1, len(cp._queue) )

  def test_publish_does_not_resend_queued_messages_if_channel_is_closed(self):
    ch = mock()
    cp = ChannelPool(None)
    user_cb = mock()
    ch.active = True
    ch.closed = False
    cp._queue.append( (('a1', 'a2'), {'cb':'foo', 'yo':'dawg'}) )

    expect(cp._get_channel).returns( ch )
    expect(ch.publish_synchronous).args( 'arg1', 'arg2', cb=var('cb'), doit='harder' )

    cp.publish( 'arg1', 'arg2', cb=user_cb, doit='harder' )
    assert_equals( set(), cp._free_channels )
    assert_equals( 1, len(cp._queue) )

    ch.closed = True
    stub(cp._process_queue)
    expect(user_cb)
    var('cb').value()
    assert_equals( set([ch]), cp._free_channels )
    assert_equals( 1, len(cp._queue) )

  def test_publish_searches_for_active_channel(self):
    ch1 = mock()
    ch2 = mock()
    ch3 = mock()
    ch1.active = ch2.active = False
    ch3.active = True
    cp = ChannelPool(None)

    expect(cp._get_channel).returns(ch1)
    expect(cp._get_channel).returns(ch2)
    expect(cp._get_channel).returns(ch3)
    expect(ch3.publish_synchronous).args( 'arg1', 'arg2', cb=ignore() )

    cp.publish( 'arg1', 'arg2' )
    self.assertEquals( set([ch1,ch2]), cp._free_channels )

  def test_publish_appends_to_queue_when_no_ready_channels(self):
    cp = ChannelPool(None)

    expect(cp._get_channel).returns(None)

    cp.publish( 'arg1', 'arg2', arg3='foo', cb='usercb' )
    self.assertEquals( set(), cp._free_channels )
    assert_equals( deque([ (('arg1','arg2'), {'arg3':'foo','cb':'usercb'})]),
      cp._queue )

  def test_publish_appends_to_queue_when_no_ready_channels_out_of_several(self):
    ch1 = mock()
    cp = ChannelPool(None)
    ch1.active = False

    expect(cp._get_channel).returns(ch1)
    expect(cp._get_channel).returns(None)

    cp.publish( 'arg1', 'arg2', arg3='foo', cb='usercb' )
    self.assertEquals( set([ch1]), cp._free_channels )
    assert_equals( deque([ (('arg1','arg2'), {'arg3':'foo','cb':'usercb'})]),
      cp._queue )

  def test_process_queue(self):
    cp = ChannelPool(None)
    cp._queue = deque([
      (('foo',),{'a':1}),
      (('bar',),{'b':2}),
    ])
    expect( cp.publish ).args('foo', a=1)
    expect( cp.publish ).args('bar', b=2)

    cp._process_queue()
    cp._process_queue()
    cp._process_queue()

  def test_get_channel_returns_new_when_none_free_and_not_at_limit(self):
    conn = mock()
    cp = ChannelPool(conn)
    cp._channels = 1

    with expect(conn.channel).returns(mock()) as newchannel:
      expect( newchannel.add_close_listener ).args( cp._channel_closed_cb )
      self.assertEquals( newchannel, cp._get_channel() )
    self.assertEquals( set(), cp._free_channels )
    assert_equals( 2, cp._channels )

  def test_get_channel_returns_new_when_none_free_and_at_limit(self):
    conn = mock()
    cp = ChannelPool(conn, 1)
    cp._channels = 1

    stub(conn.channel)

    self.assertEquals( None, cp._get_channel() )
    self.assertEquals( set(), cp._free_channels )

  def test_get_channel_when_one_free_and_not_closed(self):
    conn = mock()
    ch = mock()
    ch.closed = False
    cp = ChannelPool(conn)
    cp._free_channels = set([ch])

    self.assertEquals( ch, cp._get_channel() )
    self.assertEquals( set(), cp._free_channels )

  def test_get_channel_when_two_free_and_one_closed(self):
    # Because we can't mock builtins ....
    class Set(set):
      def pop(self): pass

    conn = mock()
    ch1 = mock()
    ch1.closed = True
    ch2 = mock()
    ch2.closed = False
    cp = ChannelPool(conn)
    cp._free_channels = Set([ch1,ch2])
    cp._channels = 2

    # Because we want them in order
    expect( cp._free_channels.pop ).returns( ch1 ).side_effect( super(Set,cp._free_channels).pop )
    expect( cp._free_channels.pop ).returns( ch2 ).side_effect( super(Set,cp._free_channels).pop )

    self.assertEquals( ch2, cp._get_channel() )
    self.assertEquals( set(), cp._free_channels )
    assert_equals( 2, cp._channels )

  def test_get_channel_when_two_free_and_all_closed(self):
    conn = mock()
    ch1 = mock()
    ch1.closed = True
    ch2 = mock()
    ch2.closed = True
    cp = ChannelPool(conn)
    cp._free_channels = set([ch1,ch2])
    cp._channels = 2

    with expect(conn.channel).returns(mock()) as newchannel:
      expect( newchannel.add_close_listener ).args( cp._channel_closed_cb )
      self.assertEquals( newchannel, cp._get_channel() )

    self.assertEquals( set(), cp._free_channels )
    assert_equals( 3, cp._channels )

  def test_channel_closed_cb(self):
    cp = ChannelPool(None)
    cp._channels = 32

    expect( cp._process_queue )
    cp._channel_closed_cb('channel')
    assert_equals( 31, cp._channels )
